# coding: utf-8
# [Pix2Text](https://github.com/breezedeus/pix2text): an Open-Source Alternative to Mathpix.
# Copyright (C) 2022-2024, [Breezedeus](https://www.breezedeus.com).

import os
import logging
import glob
import json
from multiprocessing import Process
from pprint import pformat

import click

from pix2text import set_logger, Pix2Text

_CONTEXT_SETTINGS = {"help_option_names": ['-h', '--help']}
logger = set_logger(log_level=logging.INFO)


@click.group(context_settings=_CONTEXT_SETTINGS)
def cli():
    pass


@cli.command('predict')
@click.option(
    "-l",
    "--languages",
    type=str,
    default='en,ch_sim',
    help="Language Codes for Text-OCR to recognize, separated by commas",
    show_default=True,
)
@click.option(
    "--layout-config",
    type=str,
    default=None,
    help="Configuration information for the layout parser model, in JSON string format. Default: `None`, meaning using the default configuration",
    show_default=True,
)
@click.option(
    "--mfd-config",
    type=str,
    default=None,
    help="Configuration information for the MFD model, in JSON string format. Default: `None`, meaning using the default configuration",
    show_default=True,
)
@click.option(
    "--formula-ocr-config",
    type=str,
    default=None,
    help="Configuration information for the Latex-OCR mathematical formula recognition model. Default: `None`, meaning using the default configuration",
    show_default=True,
)
@click.option(
    "--text-ocr-config",
    type=str,
    default=None,
    help="Configuration information for Text-OCR recognition, in JSON string format. Default: `None`, meaning using the default configuration",
    show_default=True,
)
@click.option(
    "--enable-formula/--disable-formula",
    default=True,
    help="Whether to enable formula recognition",
    show_default=True,
)
@click.option(
    "--enable-table/--disable-table",
    default=True,
    help="Whether to enable table recognition",
    show_default=True,
)
@click.option(
    "-d",
    "--device",
    help="Choose to run the code using `cpu`, `gpu`, or a specific GPU like `cuda:0`",
    type=str,
    default='cpu',
    show_default=True,
)
@click.option(
    "--file-type",
    type=click.Choice(['pdf', 'page', 'text_formula', 'formula', 'text']),
    default='text_formula',
    help="Which file type to process, 'pdf', 'page', 'text_formula', 'formula', or 'text'",
    show_default=True,
)
@click.option(
    "--resized-shape",
    help="Resize the image width to this size before processing",
    type=int,
    default=768,
    show_default=True,
)
@click.option(
    "-i",
    "--img-file-or-dir",
    required=True,
    help="File path of the input image/pdf or the specified directory",
)
@click.option(
    "--save-debug-res",
    default=None,
    help="If `save_debug_res` is set, the directory to save the debug results; default value is `None`, which means not to save",
    show_default=True,
)
@click.option(
    "--rec-kwargs",
    type=str,
    default=None,
    help="kwargs for calling .recognize(), in JSON string format",
    show_default=True,
)
@click.option(
    "--return-text/--no-return-text",
    default=True,
    help="Whether to return only the text result",
    show_default=True,
)
@click.option(
    "--auto-line-break/--no-auto-line-break",
    default=True,
    help="Whether to automatically determine to merge adjacent line results into a single line result",
    show_default=True,
)
@click.option(
    "-o",
    "--output-dir",
    default='output-md',
    help="Output directory for the recognized text results. Only effective when `file-type` is `pdf` or `page`",
    show_default=True,
)
@click.option(
    "--log-level",
    default='INFO',
    help="Log Level, such as `INFO`, `DEBUG`",
    show_default=True,
)
def predict(
    languages,
    layout_config,
    mfd_config,
    formula_ocr_config,
    text_ocr_config,
    enable_formula,
    enable_table,
    device,
    file_type,
    resized_shape,
    img_file_or_dir,
    save_debug_res,
    rec_kwargs,
    return_text,
    auto_line_break,
    output_dir,
    log_level,
):
    """Use Pix2Text (P2T) to predict the text information in an image or PDF."""
    logger = set_logger(log_level=log_level)

    analyzer_config = json.loads(mfd_config) if mfd_config else {}
    formula_ocr_config = json.loads(formula_ocr_config) if formula_ocr_config else {}
    languages = [lang.strip() for lang in languages.split(',') if lang.strip()]
    text_ocr_config = json.loads(text_ocr_config) if text_ocr_config else {}

    layout_config = json.loads(layout_config) if layout_config else {}
    text_formula_config = {
        'languages': languages,  # 'en,ch_sim
        'mfd': analyzer_config,
        'formula': formula_ocr_config,
        'text': text_ocr_config,
    }
    total_config = {
        'layout': layout_config,
        'text_formula': text_formula_config,
    }
    p2t = Pix2Text.from_config(
        total_configs=total_config,
        enable_formula=enable_formula,
        enable_table=enable_table,
        device=device,
    )

    fp_list = []
    if os.path.isfile(img_file_or_dir):
        fp_list.append(img_file_or_dir)
        if save_debug_res:
            save_debug_res = [save_debug_res]
    elif os.path.isdir(img_file_or_dir):
        fn_list = glob.glob1(img_file_or_dir, '*g')
        fp_list = [os.path.join(img_file_or_dir, fn) for fn in fn_list]
        if save_debug_res:
            os.makedirs(save_debug_res, exist_ok=True)
            save_debug_res = [
                os.path.join(save_debug_res, 'output-debugs-' + fn) for fn in fn_list
            ]
    else:
        raise ValueError(f'{img_file_or_dir} is not a valid file or directory')

    rec_kwargs = json.loads(rec_kwargs) if rec_kwargs else {}
    rec_kwargs['resized_shape'] = resized_shape
    rec_kwargs['return_text'] = return_text
    rec_kwargs['auto_line_break'] = auto_line_break

    for idx, fp in enumerate(fp_list):
        if file_type in ('pdf', 'page'):
            rec_kwargs['save_debug_res'] = (
                save_debug_res[idx] if save_debug_res is not None else None
            )
        else:
            rec_kwargs['save_analysis_res'] = (
                save_debug_res[idx] if save_debug_res is not None else None
            )
        out = p2t.recognize(fp, file_type=file_type, **rec_kwargs)
        if file_type in ('pdf', 'page'):
            out = out.to_markdown(output_dir)
        logger.info(
            f'In image: {fp}\nOuts: \n{out if isinstance(out, str) else pformat(out)}\n'
        )


@cli.command('serve')
@click.option(
    "-l",
    "--languages",
    type=str,
    default='en,ch_sim',
    help="Language Codes for Text-OCR to recognize, separated by commas",
    show_default=True,
)
@click.option(
    "--layout-config",
    type=str,
    default=None,
    help="Configuration information for the layout parser model, in JSON string format. Default: `None`, meaning using the default configuration",
    show_default=True,
)
@click.option(
    "--mfd-config",
    type=str,
    default=None,
    help="Configuration information for the MFD model, in JSON string format. Default: `None`, meaning using the default configuration",
    show_default=True,
)
@click.option(
    "--formula-ocr-config",
    type=str,
    default=None,
    help="Configuration information for the Latex-OCR mathematical formula recognition model. Default: `None`, meaning using the default configuration",
    show_default=True,
)
@click.option(
    "--text-ocr-config",
    type=str,
    default=None,
    help="Configuration information for Text-OCR recognition, in JSON string format. Default: `None`, meaning using the default configuration",
    show_default=True,
)
@click.option(
    "--enable-formula/--disable-formula",
    default=True,
    help="Whether to enable formula recognition",
    show_default=True,
)
@click.option(
    "--enable-table/--disable-table",
    default=True,
    help="Whether to enable table recognition",
    show_default=True,
)
@click.option(
    "-d",
    "--device",
    help="Choose to run the code using `cpu`, `gpu`, or a specific GPU like `cuda:0`",
    type=str,
    default='cpu',
    show_default=True,
)
@click.option(
    "-o",
    "--output-md-root-dir",
    default='output-md-root',
    help="Markdown output root directory for the recognized text results. Only effective when `file-type` is `pdf` or `page`",
    show_default=True,
)
@click.option(
    '-H', '--host', type=str, default='0.0.0.0', help='server host', show_default=True,
)
@click.option(
    '-p', '--port', type=int, default=8503, help='server port', show_default=True,
)
@click.option(
    '--reload',
    is_flag=True,
    help='whether to reload the server when the codes have been changed',
    show_default=True,
)
@click.option(
    "--log-level",
    default='INFO',
    help="Log Level, such as `INFO`, `DEBUG`",
    show_default=True,
)
def serve(
    languages,
    layout_config,
    mfd_config,
    formula_ocr_config,
    text_ocr_config,
    enable_formula,
    enable_table,
    device,
    output_md_root_dir,
    host,
    port,
    reload,
    log_level,
):
    """Start the HTTP service."""
    from pix2text.serve import start_server

    logger = set_logger(log_level=log_level)

    analyzer_config = json.loads(mfd_config) if mfd_config else {}
    formula_ocr_config = json.loads(formula_ocr_config) if formula_ocr_config else {}
    languages = [lang.strip() for lang in languages.split(',') if lang.strip()]
    text_ocr_config = json.loads(text_ocr_config) if text_ocr_config else {}

    layout_config = json.loads(layout_config) if layout_config else {}
    text_formula_config = {
        'languages': languages,  # 'en,ch_sim
        'mfd': analyzer_config,
        'formula': formula_ocr_config,
        'text': text_ocr_config,
    }
    total_config = {
        'layout': layout_config,
        'text_formula': text_formula_config,
    }
    p2t_config = dict(
        total_configs=total_config,
        enable_formula=enable_formula,
        enable_table=enable_table,
        device=device,
    )
    api = Process(
        target=start_server,
        kwargs={
            'p2t_config': p2t_config,
            'output_md_root_dir': output_md_root_dir,
            'host': host,
            'port': port,
            'reload': reload,
        },
    )
    api.start()
    api.join()


if __name__ == "__main__":
    cli()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import torch


def _sigmoid(x):
    y = torch.clamp(x.sigmoid_(), min=1e-4, max=1 - 1e-4)
    return y


def _gather_feat(feat, ind, mask=None):
    dim = feat.size(2)
    ind = ind.unsqueeze(2).expand(ind.size(0), ind.size(1), dim)
    if feat.device.type == 'mps' and ind.device.type == 'mps':
        feat = feat.to('cpu')
        ind = ind.to('cpu')
        feat = feat.gather(1, ind)
        feat = feat.to('mps')
    else:
        feat = feat.gather(1, ind)
    if mask is not None:
        mask = mask.unsqueeze(2).expand_as(feat)
        feat = feat[mask]
        feat = feat.view(-1, dim)
    return feat


def _tranpose_and_gather_feat(feat, ind):
    feat = feat.permute(0, 2, 3, 1).contiguous()
    feat = feat.view(feat.size(0), -1, feat.size(3))
    feat = _gather_feat(feat, ind)
    return feat


def flip_tensor(x):
    return torch.flip(x, [3])
    # tmp = x.detach().cpu().numpy()[..., ::-1].copy()
    # return torch.from_numpy(tmp).to(x.device)


def flip_lr(x, flip_idx):
    tmp = x.detach().cpu().numpy()[..., ::-1].copy()
    shape = tmp.shape
    for e in flip_idx:
        tmp[:, e[0], ...], tmp[:, e[1], ...] = \
            tmp[:, e[1], ...].copy(), tmp[:, e[0], ...].copy()
    return torch.from_numpy(tmp.reshape(shape)).to(x.device)


def flip_lr_off(x, flip_idx):
    tmp = x.detach().cpu().numpy()[..., ::-1].copy()
    shape = tmp.shape
    tmp = tmp.reshape(tmp.shape[0], 17, 2,
                      tmp.shape[2], tmp.shape[3])
    tmp[:, :, 0, :, :] *= -1
    for e in flip_idx:
        tmp[:, e[0], ...], tmp[:, e[1], ...] = \
            tmp[:, e[1], ...].copy(), tmp[:, e[0], ...].copy()
    return torch.from_numpy(tmp.reshape(shape)).to(x.device)



# coding: utf-8
import random
from pprint import pprint
import numpy as np

from pix2text.utils import sort_and_filter_line_boxes, sort_boxes


def list2box(xmin, ymin, xmax, ymax):
    return np.array([[xmin, ymin], [xmax, ymin], [xmax, ymax], [xmin, ymax]])


def test_sort_line_boxes():
    boxes = [
        {'position': list2box(20, 0, 30, 25), 'id': 12},  # anchor，不能动
        {'position': list2box(0, 5, 20, 20), 'id': 11},
        {'position': list2box(30, 5, 40, 18), 'id': 13},
        {'position': list2box(38, 5, 60, 20), 'id': 14},
        {'position': list2box(21, 20, 40, 30), 'id': 22},
    ]
    for box in boxes:
        box['__line__'] = -1  # 所在行号，-1表示未分配
    boxes[0]['__line__'] = 1
    # random.shuffle(boxes)
    outs = sort_and_filter_line_boxes(boxes, 'position')
    pprint(outs)


def test_sort_boxes():
    boxes = [
        {'position': list2box(0, 5, 20, 20), 'id': 11},
        {'position': list2box(30, 5, 40, 18), 'id': 13},
        {'position': list2box(38, 5, 60, 20), 'id': 14},
        {'position': list2box(20, 0, 30, 25), 'id': 12},
        {'position': list2box(0, 25, 20, 45), 'id': 21},
        {'position': list2box(40, 25, 60, 44), 'id': 23},
        {'position': list2box(21, 20, 40, 30), 'id': 22},
        {'position': list2box(2, 85, 30, 105), 'id': 41},
        {'position': list2box(10, 55, 50, 80), 'id': 31},
        {'position': list2box(35, 83, 58, 103), 'id': 42},
    ]

    line_boxes = sort_boxes(boxes, 'position')
    pprint(line_boxes)

    line_ids = []
    for boxes in line_boxes:
        line_ids.append([box['id'] for box in boxes])

    pprint(line_ids)
    assert line_ids == [
        [11, 12, 13, 14],
        [21, 22, 23],
        [31],
        [41, 42],
    ]



# coding: utf-8

from typing import List, Dict, Any

import numpy as np
import torch
from torchvision.utils import draw_bounding_boxes
from torchvision.transforms.functional import to_pil_image
from torchvision.models.detection import (
    fcos_resnet50_fpn,
    FCOS_ResNet50_FPN_Weights,
    fasterrcnn_resnet50_fpn_v2,
    FasterRCNN_ResNet50_FPN_V2_Weights,
    fasterrcnn_mobilenet_v3_large_fpn,
    FasterRCNN_MobileNet_V3_Large_FPN_Weights,
    retinanet_resnet50_fpn_v2,
    RetinaNet_ResNet50_FPN_V2_Weights,
    ssd300_vgg16,
    SSD300_VGG16_Weights,
    ssdlite320_mobilenet_v3_large,
    SSDLite320_MobileNet_V3_Large_Weights,
)

from .utils import read_img
from .category_mapping import CATEGORY_MAPPINGS


MODELS = {
    "frcnn-resnet": (
        fasterrcnn_resnet50_fpn_v2,
        FasterRCNN_ResNet50_FPN_V2_Weights.DEFAULT,
    ),  # 7s
    "frcnn-mobilenet": (  # 1s
        fasterrcnn_mobilenet_v3_large_fpn,
        FasterRCNN_MobileNet_V3_Large_FPN_Weights.DEFAULT,
    ),
    "retinanet": (
        retinanet_resnet50_fpn_v2,
        RetinaNet_ResNet50_FPN_V2_Weights.DEFAULT,
    ),  # 3s
    "ssd300_vgg16": (ssd300_vgg16, SSD300_VGG16_Weights.DEFAULT),
    "ssdlite320_mobilenet_v3_large": (  # <1s
        ssdlite320_mobilenet_v3_large,
        SSDLite320_MobileNet_V3_Large_Weights.DEFAULT,
    ),
    "fcos_resnet50_fpn": (fcos_resnet50_fpn, FCOS_ResNet50_FPN_Weights.DEFAULT),  # 3s
}


class ObjectDetector(object):
    def __init__(
        self, model_name='frcnn-mobilenet', score_thresh=0.6, nms_thresh=0.2, **kwargs
    ):
        model_cls, weights = MODELS[model_name]
        cls_configs = dict(
            score_thresh=score_thresh,
            box_score_thresh=score_thresh,
            nms_thresh=nms_thresh,
            box_nms_thresh=nms_thresh,
        )
        self.model = model_cls(weights=weights, **cls_configs)
        self.model.eval()
        self.transform = weights.transforms()
        self.categories = weights.meta["categories"]

    @torch.no_grad()
    def __call__(self, img) -> List[Dict[str, Any]]:
        batch = [self.transform(img)]
        prediction = self.model(batch)[0]
        labels = [
            CATEGORY_MAPPINGS.get(self.categories[i], self.categories[i])
            for i in prediction["labels"]
        ]
        results = []
        for l, s, b in zip(labels, prediction['scores'], prediction['boxes']):
            results.append(({'text': l, 'score': float(s), 'position': b.to( 'int')}))

        return results


if __name__ == '__main__':
    import time
    import matplotlib.pyplot as plt

    # import requests
    # response = requests.get('https://i.ytimg.com/vi/q71MCWAEfL8/maxresdefault.jpg')
    # open("obj_det.jpeg", "wb").write(response.content)
    img = read_img("docs/examples/00810-good-0.8948.jpg")
    detector = ObjectDetector(model_name='frcnn-mobilenet')

    start_time = time.time()
    out = detector(img)
    print(f'time cost: {time.time() - start_time}')

    print(out)
    labels = [_one['text'] for _one in out]
    boxes = [_one['position'] for _one in out]

    colors = (
        np.random.uniform(0, 255, size=(len(detector.categories), 3))
        .astype('int')
        .tolist()
    )
    colors = [tuple(c) for c in colors]

    box = draw_bounding_boxes(
        img,
        boxes=torch.stack(boxes, dim=0),
        labels=labels,
        colors=colors,
        width=2,
        font_size=30,
        font='Arial',
    )

    im = to_pil_image(box.detach())

    fig, ax = plt.subplots(figsize=(16, 12))
    ax.imshow(im)
    plt.show()



# coding: utf-8
# Copyright (C) 2022, [Breezedeus](https://github.com/breezedeus).

from PIL import Image
import streamlit as st

from pix2text import set_logger, Pix2Text

logger = set_logger()
st.set_page_config(layout="wide")


@st.cache(allow_output_mutation=True)
def get_model():
    return Pix2Text()


def main():
    p2t = get_model()

    title = '开源工具 <a href="https://github.com/breezedeus/pix2text">Pix2Text</a> Demo'
    st.markdown(f"<h1 style='text-align: center;'>{title}</h1>", unsafe_allow_html=True)

    subtitle = '作者：<a href="https://github.com/breezedeus">breezedeus</a>； ' \
               '欢迎加入 <a href="https://cnocr.readthedocs.io/zh/latest/contact/">交流群</a>'

    st.markdown(f"<div style='text-align: center;'>{subtitle}</div>", unsafe_allow_html=True)
    st.markdown('')
    st.subheader('选择待识别图片')
    content_file = st.file_uploader('', type=["png", "jpg", "jpeg", "webp"])
    if content_file is None:
        st.stop()

    try:
        img = Image.open(content_file).convert('RGB')
        img.save('ori.jpg')

        out = p2t(img)
        logger.info(out)
        st.markdown('##### 原始图片：')
        cols = st.columns([1, 3, 1])
        with cols[1]:
            st.image(content_file)

        st.subheader('识别结果：')
        st.markdown(f"* **图片类型**：{out['image_type']}")
        st.markdown("* **识别内容**：")

        cols = st.columns([1, 3, 1])
        with cols[1]:
            st.text(out['text'])

            if out['image_type'] == 'formula':
                st.markdown(f"$${out['text']}$$")

    except Exception as e:
        st.error(e)


if __name__ == '__main__':
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from .ddd_utils import ddd2locrot
from .image import transform_preds


def get_pred_depth(depth):
    return depth


def get_alpha(rot):
    # output: (B, 8) [bin1_cls[0], bin1_cls[1], bin1_sin, bin1_cos,
    #                 bin2_cls[0], bin2_cls[1], bin2_sin, bin2_cos]
    # return rot[:, 0]
    idx = rot[:, 1] > rot[:, 5]
    alpha1 = np.arctan(rot[:, 2] / rot[:, 3]) + (-0.5 * np.pi)
    alpha2 = np.arctan(rot[:, 6] / rot[:, 7]) + (0.5 * np.pi)
    return alpha1 * idx + alpha2 * (1 - idx)


def ddd_post_process_2d(dets, c, s, opt):
    # dets: batch x max_dets x dim
    # return 1-based class det list
    ret = []
    include_wh = dets.shape[2] > 16
    for i in range(dets.shape[0]):
        top_preds = {}
        dets[i, :, :2] = transform_preds(
            dets[i, :, 0:2], c[i], s[i], (opt.output_w, opt.output_h))
        classes = dets[i, :, -1]
        for j in range(opt.num_classes):
            inds = (classes == j)
            top_preds[j + 1] = np.concatenate([
                dets[i, inds, :3].astype(np.float32),
                get_alpha(dets[i, inds, 3:11])[:, np.newaxis].astype(np.float32),
                get_pred_depth(dets[i, inds, 11:12]).astype(np.float32),
                dets[i, inds, 12:15].astype(np.float32)], axis=1)
            if include_wh:
                top_preds[j + 1] = np.concatenate([
                    top_preds[j + 1],
                    transform_preds(
                        dets[i, inds, 15:17], c[i], s[i], (opt.output_w, opt.output_h))
                        .astype(np.float32)], axis=1)
        ret.append(top_preds)
    return ret


def ddd_post_process_3d(dets, calibs):
    # dets: batch x max_dets x dim
    # return 1-based class det list
    ret = []
    for i in range(len(dets)):
        preds = {}
        for cls_ind in dets[i].keys():
            preds[cls_ind] = []
            for j in range(len(dets[i][cls_ind])):
                center = dets[i][cls_ind][j][:2]
                score = dets[i][cls_ind][j][2]
                alpha = dets[i][cls_ind][j][3]
                depth = dets[i][cls_ind][j][4]
                dimensions = dets[i][cls_ind][j][5:8]
                wh = dets[i][cls_ind][j][8:10]
                locations, rotation_y = ddd2locrot(
                    center, alpha, dimensions, depth, calibs[0])
                bbox = [center[0] - wh[0] / 2, center[1] - wh[1] / 2,
                        center[0] + wh[0] / 2, center[1] + wh[1] / 2]
                pred = [alpha] + bbox + dimensions.tolist() + \
                       locations.tolist() + [rotation_y, score]
                preds[cls_ind].append(pred)
            preds[cls_ind] = np.array(preds[cls_ind], dtype=np.float32)
        ret.append(preds)
    return ret


def ddd_post_process(dets, c, s, calibs, opt):
    # dets: batch x max_dets x dim
    # return 1-based class det list
    dets = ddd_post_process_2d(dets, c, s, opt)
    dets = ddd_post_process_3d(dets, calibs)
    return dets


def ctdet_4ps_post_process(dets, c, s, h, w, num_classes):
    # dets: batch x max_dets x dim
    # return 1-based class det dict
    ret = []
    for i in range(dets.shape[0]):
        top_preds = {}
        dets[i, :, 0:2] = transform_preds(dets[i, :, 0:2], c[i], s[i], (w, h))
        dets[i, :, 2:4] = transform_preds(dets[i, :, 2:4], c[i], s[i], (w, h))
        dets[i, :, 4:6] = transform_preds(dets[i, :, 4:6], c[i], s[i], (w, h))
        dets[i, :, 6:8] = transform_preds(dets[i, :, 6:8], c[i], s[i], (w, h))
        classes = dets[i, :, 9]
        for j in range(num_classes):
            inds = (classes == j)
            top_preds[j + 1] = np.concatenate([
                dets[i, inds, :8].astype(np.float32),
                dets[i, inds, 8:].astype(np.float32)], axis=1).tolist()
        ret.append(top_preds)
    return ret


def ctdet_post_process(dets, c, s, h, w, num_classes):
    # dets: batch x max_dets x dim
    # return 1-based class det dict
    ret = []
    for i in range(dets.shape[0]):
        top_preds = {}
        dets[i, :, :2] = transform_preds(
            dets[i, :, 0:2], c[i], s[i], (w, h))
        dets[i, :, 2:4] = transform_preds(
            dets[i, :, 2:4], c[i], s[i], (w, h))
        classes = dets[i, :, -1]
        for j in range(num_classes):
            inds = (classes == j)
            top_preds[j + 1] = np.concatenate([
                dets[i, inds, :4].astype(np.float32),
                dets[i, inds, 4:5].astype(np.float32)], axis=1).tolist()
        ret.append(top_preds)
    return ret


def ctdet_corner_post_process(corner, c, s, h, w, num_classes):
    corner[:, :2] = transform_preds(corner[:, 0:2], c[0], s[0], (w, h))
    return corner


def multi_pose_post_process(dets, c, s, h, w):
    # dets: batch x max_dets x 40
    # return list of 39 in image coord
    ret = []
    for i in range(dets.shape[0]):
        bbox = transform_preds(dets[i, :, :4].reshape(-1, 2), c[i], s[i], (w, h))
        pts = transform_preds(dets[i, :, 5:39].reshape(-1, 2), c[i], s[i], (w, h))
        top_preds = np.concatenate(
            [bbox.reshape(-1, 4), dets[i, :, 4:5],
             pts.reshape(-1, 34)], axis=1).astype(np.float32).tolist()
        ret.append({np.ones(1, dtype=np.int32)[0]: top_preds})
    return ret



# coding: utf-8
# Copyright (C) 2022, [Breezedeus](https://github.com/breezedeus).

# 安装 pyperclip
# > pip install pyperclip

import os
import time
import glob

import pyperclip as pc

from pix2text import set_logger, Pix2Text, merge_line_texts, render_html

logger = set_logger(log_level='DEBUG')

SCREENSHOT_DIR = os.getenv(
    "SCREENSHOT_DIR", '/Users/king/Pictures/screenshot_from_xnip'
)

thresholds = {
    'formula2general': 0.65,  # 如果识别为 `formula` 类型，但得分小于此阈值，则改为 `general` 类型
    'english2general': 0.75,  # 如果识别为 `english` 类型，但得分小于此阈值，则改为 `general` 类型
}
config = dict(analyzer=dict(model_name='mfd'), thresholds=thresholds)
P2T = Pix2Text.from_config(config)


def get_newest_fp_time(screenshot_dir):
    fn_list = glob.glob1(screenshot_dir, '*g')
    fp_list = [os.path.join(screenshot_dir, fn) for fn in fn_list]
    if not fp_list:
        return None, None
    fp_list.sort(key=lambda fp: os.path.getmtime(fp), reverse=True)
    return fp_list[0], os.path.getmtime(fp_list[0])


def recognize(screenshot_dir, delta_interval):
    while True:
        newest_fp, newest_mod_time = get_newest_fp_time(screenshot_dir)
        if (
            newest_mod_time is not None
            and time.time() - newest_mod_time < delta_interval
        ):
            logger.info(f'analyzing screenshot file {newest_fp}')
            image_type, result = _recognize_newest(newest_fp)
            logger.info('image type: %s, image text: %s', image_type, result)
            if result:
                pc.copy(result)
            # render_html('./analysis_res.jpg', image_type, result, out_html_fp='out-text.html')
        time.sleep(1)


def _recognize_newest(newest_fp):
    res = P2T.recognize(
        newest_fp,
        use_analyzer=True,
        save_analysis_res='./analysis_res.jpg',
        embed_sep=(' $$', '$$ '),
        isolated_sep=('\n', '\n'),
    )
    if len(res) == 1:
        return res[0]['type'], res[0]['text']
    elif len(res) > 1:
        box_types = set([info['type'] for info in res])
        if len(box_types) > 1:
            image_type = 'hybrid'
        else:
            image_type = list(box_types)[0]
        text = merge_line_texts(res, auto_line_break=True)

        return image_type, text

    return 'general', ''


if __name__ == '__main__':
    recognize(SCREENSHOT_DIR, 1.05)



# coding: utf-8
# [Pix2Text](https://github.com/breezedeus/pix2text): an Open-Source Alternative to Mathpix.
# Copyright (C) 2022-2024, [Breezedeus](https://www.breezedeus.com).
import time
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Any, Union, Optional

from pydantic import BaseModel
from fastapi import FastAPI, UploadFile, Form, HTTPException

from pix2text import set_logger, read_img, Pix2Text

logger = set_logger(log_level='DEBUG')

app = FastAPI()


@app.get("/")
async def root():
    return {"message": "Welcome to Pix2Text Server!"}


class Pix2TextResponse(BaseModel):
    status_code: int = 200
    results: Union[str, List[Dict[str, Any]]]
    output_dir: Optional[str] = None

    def dict(self, **kwargs):
        the_dict = deepcopy(super().dict())
        return the_dict


@app.post("/pix2text")
async def ocr(
    image: UploadFile,
    file_type: str = Form(default='text_formula'),
    resized_shape: str = Form(default=768),
    embed_sep: str = Form(default=' $,$ '),
    isolated_sep: str = Form(default='$$\n, \n$$'),
) -> Dict[str, Any]:
    # curl 调用方式：
    # $ curl -F image=@docs/examples/english.jpg --form 'image_type=mixed' --form 'resized_shape=768' \
    #       http://0.0.0.0:8503/pix2text
    global P2T, OUTPUT_MD_ROOT_DIR
    if file_type not in ('text', 'formula', 'text_formula', 'page'):
        raise HTTPException(status_code=400, detail='file_type must be one of "text", "formula", "text_formula", "page"')

    img_file = image.file
    fn = Path(image.filename)
    img0 = read_img(img_file, return_type='Image')
    embed_sep = embed_sep.split(',')
    isolated_sep = isolated_sep.split(',')
    # use_analyzer = use_analyzer.lower() != 'false' if isinstance(use_analyzer, str) else use_analyzer

    params = dict(resized_shape=int(resized_shape), return_text=True)
    if len(embed_sep) == 2:
        params['embed_sep'] = embed_sep
    if len(isolated_sep) == 2:
        params['isolated_sep'] = isolated_sep

    logger.info(f'input {params=}')

    res = P2T.recognize(img0, file_type=file_type, **params)
    output_dir = None
    if file_type in ('pdf', 'page'):
        output_dir = str(OUTPUT_MD_ROOT_DIR / f'{fn.stem}-{time.time()}')
        res = res.to_markdown(output_dir)
    logger.info(f'output {res=}')

    return Pix2TextResponse(results=res, output_dir=output_dir).dict()


def start_server(
    p2t_config, output_md_root_dir, host='0.0.0.0', port=8503, reload=False, **kwargs
):
    global P2T, OUTPUT_MD_ROOT_DIR
    OUTPUT_MD_ROOT_DIR = Path(output_md_root_dir)
    OUTPUT_MD_ROOT_DIR.mkdir(exist_ok=True, parents=True)
    P2T = Pix2Text.from_config(**p2t_config)
    import uvicorn

    uvicorn.run(app, host=host, port=port, reload=reload, **kwargs)



# coding: utf-8
# Adapted from https://github.com/AlibabaResearch/AdvancedLiterateMachinery
import json
import os
import shutil
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
import logging
from typing import Union, List, Dict, Any, Optional

import numpy as np
from PIL import Image

from .opts import opts
from .huntie_subfield import Huntie_Subfield
from .detectors.detector_factory import detector_factory
from .wrapper import wrap_result
from ..consts import MODEL_VERSION
from ..layout_parser import LayoutParser, ElementType
from ..utils import (
    select_device,
    read_img,
    data_dir,
    save_layout_img,
    clipbox,
    overlap,
    box2list,
    x_overlap,
    merge_boxes,
)

logger = logging.getLogger(__name__)

CATEGORIES = {
    "title": 0,
    "figure": 1,
    "plain text": 2,
    "header": 3,
    "page number": 4,
    "footnote": 5,
    "footer": 6,
    "table": 7,
    "table caption": 8,
    "figure caption": 9,
    "equation": 10,
    "full column": 11,
    "sub column": 12,
}
CATEGORY_MAPPING = [''] * len(CATEGORIES)
for cate, idx in CATEGORIES.items():
    CATEGORY_MAPPING[idx] = cate


class DocXLayoutOutput:
    def __init__(self, layout_detection_info, subfield_detection_info, message=''):
        self.layout_detection_info = layout_detection_info
        self.subfield_detection_info = subfield_detection_info
        self.message = message

    def to_json(self):
        return wrap_result(
            self.layout_detection_info, self.subfield_detection_info, CATEGORY_MAPPING
        )


class DocXLayoutParser(LayoutParser):
    ignored_types = {'footnote', 'footer', 'page number'}
    type_mappings = {
        'title': ElementType.TITLE,
        'figure': ElementType.FIGURE,
        'plain text': ElementType.TEXT,
        'header': ElementType.TEXT,
        'table': ElementType.TABLE,
        'table caption': ElementType.TEXT,
        'figure caption': ElementType.TEXT,
        'equation': ElementType.FORMULA,
    }
    # types that are isolated and usually don't cross different columns. They should not be merged with other elements
    is_isolated = {'header', 'table caption', 'figure caption', 'equation'}

    def __init__(
        self,
        device: str = None,
        model_fp: Optional[str] = None,
        root: Union[str, Path] = data_dir(),
        **kwargs,
    ):
        if model_fp is None:
            model_fp = self._prepare_model_files(root, None)
        new_params = {
            'task': 'ctdet_subfield',
            'arch': 'dlav0subfield_34',
            'input_res': 768,
            'num_classes': 13,
            'scores_thresh': kwargs.get('scores_thresh', 0.35),
            'load_model': str(model_fp),
            'debug': kwargs.get('debug', 0),
        }

        opt = opts().parse(new_params)
        opt = opts().update_dataset_info_and_set_heads(opt, Huntie_Subfield)
        opt.device = select_device(device)

        Detector = detector_factory[opt.task]
        detector = Detector(opt)
        self.detector = detector
        self.opt = opt
        logger.debug("DocXLayoutParser parameters %s", self.opt)

    @classmethod
    def from_config(cls, configs: Optional[dict] = None, device: str = None, **kwargs):
        configs = configs or {}
        device = select_device(device)
        # configs['device'] = device if device != 'mps' else 'cpu'

        return cls(
            device=device,
            model_fp=configs.get('model_fp', None),
            root=configs.get('root', data_dir()),
            **configs,
        )

    def _prepare_model_files(self, root, model_info):
        model_root_dir = Path(root) / MODEL_VERSION
        model_dir = model_root_dir / 'layout-parser'
        model_fp = model_dir / 'DocXLayout_231012.pth'
        if model_fp.exists():
            return model_fp
        if model_dir.exists():
            shutil.rmtree(str(model_dir))
        model_dir.mkdir(parents=True)
        download_cmd = f'huggingface-cli download --repo-type model --resume-download --local-dir-use-symlinks False breezedeus/pix2text-layout --local-dir {model_dir}'
        os.system(download_cmd)
        if not model_fp.exists():  # download failed above
            if model_dir.exists():
                shutil.rmtree(str(model_dir))
            os.system('HF_ENDPOINT=https://hf-mirror.com ' + download_cmd)
        return model_fp

    def convert_eval_format(self, all_bboxes, opt):
        layout_detection_items = []
        subfield_detection_items = []
        for cls_ind in all_bboxes:
            for box in all_bboxes[cls_ind]:
                if box[8] < opt.scores_thresh:
                    continue
                pts = np.round(box).tolist()[:8]
                score = box[8]
                category_id = box[9]
                # direction_id = box[10]
                # secondary_id = box[11]
                detection = {
                    "category_id": int(category_id),
                    # "secondary_id": int(secondary_id),
                    # "direction_id": int(direction_id),
                    "poly": pts,
                    "score": float("{:.2f}".format(score)),
                }
                if cls_ind in (12, 13):
                    subfield_detection_items.append(detection)
                else:
                    layout_detection_items.append(detection)
        return layout_detection_items, subfield_detection_items

    def parse(
        self,
        img: Union[str, Path, Image.Image],
        table_as_image: bool = False,
        **kwargs,
    ) -> (List[Dict[str, Any]], Dict[str, Any]):
        """

        Args:
            img ():
            table_as_image ():
            **kwargs ():
              * save_debug_res (str): if `save_debug_res` is set, the directory to save the debug results; default value is `None`, which means not to save
              * expansion_margin (int): expansion margin

        Returns:

        """
        if isinstance(img, Image.Image):
            img0 = img.convert('RGB')
        else:
            img0 = read_img(img, return_type='Image')
        img_width, img_height = img0.size
        try:
            # to np.array, RGB -> BGR
            ret = self.detector.run(np.array(img0)[:, :, ::-1])
            layout_detection_info, subfield_detection_info = self.convert_eval_format(
                ret['results'], self.opt
            )
            out = DocXLayoutOutput(
                layout_detection_info, subfield_detection_info, message='success'
            )
        except Exception as e:
            logger.warning("DocXLayoutPredictor Error %s", repr(e))
            out = DocXLayoutOutput([], [], message=repr(e))

        layout_out = out.to_json()
        debug_dir = None
        if kwargs.get('save_debug_res', None):
            debug_dir = Path(kwargs.get('save_debug_res'))
            debug_dir.mkdir(exist_ok=True, parents=True)
        if debug_dir is not None:
            with open(debug_dir / 'layout_out.json', 'w', encoding='utf-8') as f:
                json.dump(
                    layout_out, f, indent=2, ensure_ascii=False,
                )
        if layout_out:
            layout_out = self._preprocess_outputs(img0, layout_out)
            layout_out, column_meta = self._format_outputs(
                img0, layout_out, table_as_image
            )
        else:
            layout_out, column_meta = [], {}

        layout_out = self._merge_overlapped_boxes(layout_out)

        expansion_margin = kwargs.get('expansion_margin', 8)
        layout_out = self._expand_boxes(
            layout_out, expansion_margin, height=img_height, width=img_width
        )

        save_layout_fp = kwargs.get(
            'save_layout_res',
            debug_dir / 'layout_res.jpg' if debug_dir is not None else None,
        )
        if save_layout_fp:
            element_type_list = [t for t in ElementType]
            save_layout_img(
                img0,
                element_type_list,
                layout_out,
                save_path=save_layout_fp,
                key='position',
            )

        return layout_out, column_meta

    def _preprocess_outputs(self, img0, outs):
        width, height = img0.size

        subfields = outs['subfields']
        for column_info in subfields:
            layout_out = column_info['layouts']
            if len(layout_out) < 2:
                continue
            for idx, cur_box_info in enumerate(layout_out[:-1]):
                next_box_info = layout_out[idx + 1]
                cur_box_ymax = cur_box_info['pts'][-1]
                next_box_ymin = next_box_info['pts'][1]
                if (
                    cur_box_info['category'] == 'figure'
                    and next_box_info['category'] == 'figure caption'
                    and -6 < next_box_ymin - cur_box_ymax < 80
                ):
                    new_xmin = min(cur_box_info['pts'][0], next_box_info['pts'][0])
                    # new_xmin = max(new_xmin, 0, col_pts[0])
                    new_xmax = max(cur_box_info['pts'][2], next_box_info['pts'][2])
                    # new_xmax = min(new_xmax, )
                    new_ymin = max(0, cur_box_info['pts'][1])
                    new_ymax = max(cur_box_ymax, next_box_ymin - 16)
                    new_box = [
                        new_xmin,
                        new_ymin,
                        new_xmax,
                        new_ymin,
                        new_xmax,
                        new_ymax,
                        new_xmin,
                        new_ymax,
                    ]
                    layout_out[idx]['pts'] = new_box
                # FIXME: first figure caption, then figure

        return outs

    def _format_outputs(self, img0, out, table_as_image: bool):
        width, height = img0.size

        column_meta = defaultdict(dict)
        final_out = []
        subfields = out['subfields']
        col_number = 0
        for column_info in subfields:
            if column_info['category'] == 'sub column':
                cur_col_number = col_number
                col_number += 1
            elif column_info['category'] == 'full column':  # == 'full column'
                cur_col_number = -1
            else:  # '其他'
                cur_col_number = -2
            box = clipbox(np.array(column_info['pts']).reshape(4, 2), height, width)
            column_meta[cur_col_number]['position'] = box
            column_meta[cur_col_number]['score'] = column_info['confidence']
            layout_out = column_info['layouts']
            for box_info in layout_out:
                image_type = box_info['category']
                isolated = image_type in self.is_isolated
                if image_type in self.ignored_types:
                    image_type = ElementType.IGNORED
                else:
                    image_type = self.type_mappings.get(image_type, ElementType.UNKNOWN)
                if table_as_image and image_type == ElementType.TABLE:
                    image_type = ElementType.FIGURE
                box = clipbox(np.array(box_info['pts']).reshape(4, 2), height, width)
                final_out.append(
                    {
                        'type': image_type,
                        'position': box,
                        'score': box_info['confidence'],
                        'col_number': cur_col_number,
                        'isolated': isolated,
                    }
                )

        if -2 in column_meta and -1 in column_meta:
            filtered_out = []
            full_column_box = column_meta[-1]['position']
            full_column_xmin, full_column_xmax = (
                full_column_box[0, 0],
                full_column_box[1, 0],
            )
            for box_info in final_out:
                if box_info['col_number'] != -2:
                    filtered_out.append(box_info)
                    continue
                cur_box = box_info['position']
                cur_box_xmin, cur_box_xmax = cur_box[0, 0], cur_box[1, 0]
                cur_box_ymin, cur_box_ymax = cur_box[0, 1], cur_box[2, 1]
                if (
                    box_info['type'] == ElementType.TEXT
                    and (
                        cur_box_xmax < full_column_xmin
                        or cur_box_xmin > full_column_xmax
                    )
                    and cur_box_ymax - cur_box_ymin > 5 * (cur_box_xmax - cur_box_xmin)
                ):  # unnecessary block
                    box_info['type'] = ElementType.IGNORED
                filtered_out.append(box_info)

            final_out = filtered_out

        # handle abnormal elements (col_number == -2)
        if -2 in column_meta:
            column_meta.pop(-2)
        # guess which column the box belongs to
        for _box_info in final_out:
            if _box_info['col_number'] != -2:
                continue
            overlap_vals = []
            for col_number, col_info in column_meta.items():
                overlap_val = x_overlap(_box_info, col_info, key='position')
                overlap_vals.append([col_number, overlap_val])
            overlap_vals.sort(key=lambda x: (x[1], x[0]), reverse=True)
            match_col_number = overlap_vals[0][0]
            _box_info['col_number'] = match_col_number

        return final_out, column_meta

    def _merge_overlapped_boxes(self, layout_out):
        """
        Detected bounding boxes may overlap; merge these overlapping boxes into a single one.
        """
        if len(layout_out) < 2:
            return layout_out
        layout_out = deepcopy(layout_out)

        def _overlay_vertically(box1, box2):
            if x_overlap(box1, box2, key=None) < 0.8:
                return False
            box1 = box2list(box1)
            box2 = box2list(box2)
            # 判断是否有交集
            if box1[3] <= box2[1] or box2[3] <= box1[1]:
                return False
            # 计算交集的高度
            y_min = max(box1[1], box2[1])
            y_max = min(box1[3], box2[3])
            return y_max - y_min > 10

        for anchor_idx, anchor_box_info in enumerate(layout_out):
            if anchor_box_info['type'] != ElementType.TEXT or anchor_box_info.get(
                'used', False
            ):
                continue
            for cand_idx, cand_box_info in enumerate(layout_out):
                if anchor_idx == cand_idx:
                    continue
                if cand_box_info['type'] != ElementType.TEXT or cand_box_info.get(
                    'used', False
                ):
                    continue
                if not _overlay_vertically(
                    anchor_box_info['position'], cand_box_info['position']
                ):
                    continue
                anchor_box_info['position'] = merge_boxes(
                    anchor_box_info['position'], cand_box_info['position']
                )
                cand_box_info['used'] = True

        return [box_info for box_info in layout_out if not box_info.get('used', False)]

    def _expand_boxes(self, layout_out, expansion_margin, height, width):
        """
        Expand boxes with some margin to get better results
        Args:
            layout_out (): layout_out
            expansion_margin (int): expansion margin
            height (int): height of the image
            width (int): width of the image

        Returns: layout_out with expanded boxes

        """

        def _overlap_with_some_box(idx, anchor_box):
            # anchor_box = layout_out[idx]
            return any(
                [
                    overlap(anchor_box, box_info['position'], key=None) > 0
                    for idx2, box_info in enumerate(layout_out)
                    if idx2 != idx
                ]
            )

        for idx, box_info in enumerate(layout_out):
            if box_info['type'] not in (
                ElementType.TEXT,
                ElementType.TITLE,
                ElementType.FORMULA,
            ):
                continue
            if _overlap_with_some_box(idx, box_info['position']):
                continue

            # expand xmin and xmax
            new_box = box_info['position'].copy()
            xmin, xmax = new_box[0, 0], new_box[1, 0]
            xmin -= expansion_margin
            xmax += expansion_margin
            if xmin <= 8:
                xmin = 0
            if xmax + 8 >= width:
                xmax = width
            new_box[0, 0] = new_box[3, 0] = xmin
            new_box = clipbox(new_box, height, width)
            if not _overlap_with_some_box(idx, new_box):
                layout_out[idx]['position'] = new_box
            new_box = layout_out[idx]['position'].copy()
            new_box[1, 0] = new_box[2, 0] = xmax
            new_box = clipbox(new_box, height, width)
            if not _overlap_with_some_box(idx, new_box):
                layout_out[idx]['position'] = new_box

            # expand ymin and ymax
            new_box = layout_out[idx]['position'].copy()
            ymin, ymax = new_box[0, 1], new_box[2, 1]
            ymin -= expansion_margin
            ymax += expansion_margin
            if ymin <= 8:
                ymin = 0
            if ymax + 8 >= height:
                ymax = height
            new_box[0, 1] = new_box[1, 1] = ymin
            new_box = clipbox(new_box, height, width)
            if not _overlap_with_some_box(idx, new_box):
                layout_out[idx]['position'] = new_box
            new_box = layout_out[idx]['position'].copy()
            new_box[2, 1] = new_box[3, 1] = ymax
            new_box = clipbox(new_box, height, width)
            if not _overlap_with_some_box(idx, new_box):
                layout_out[idx]['position'] = new_box

        return layout_out



# coding: utf-8

from PIL import Image
import layoutparser as lp

# config_path = 'lp://PubLayNet/faster_rcnn_R_50_FPN_3x/config'  # In model catalog
# label_map = {
#     0: "Text",
#     1: "Title",
#     2: "List",
#     3: "Table",
#     4: "Figure",
# }  # In model`label_map`

# model = lp.Detectron2LayoutModel(
#     config_path=config_path,
#     label_map=label_map,
#     extra_config=["MODEL.ROI_HEADS.SCORE_THRESH_TEST", 0.8],  # Optional
# )

model = lp.AutoLayoutModel(
    "lp://efficientdet/MFD/tf_efficientdet_d1",
    model_path='mfd-tf_efficientdet_d1.pth.tar',
    extra_config={"output_confidence_threshold": 0.5},
)


image = Image.open('docs/examples/layout3.jpg')

out = model.detect(image)
print(out)

lp.draw_box(image, out, box_width=3).show()
# breakpoint()



# coding: utf-8
# Copyright (C) 2022-2024, [Breezedeus](https://github.com/breezedeus).
#
# credit to: pix2tex, lukas-blecher/LaTeX-OCR
# Adapted from https://github.com/lukas-blecher/LaTeX-OCR/blob/main/pix2tex/cli.py
from typing import Tuple, Optional, Dict, Any, Union
import logging
import yaml
from pathlib import Path

from PIL import Image
from transformers import PreTrainedTokenizerFast
from timm.models.resnetv2 import ResNetV2
from timm.models.layers import StdConv2dSame
from cnstd.utils import get_model_file

from pix2tex.utils import *
from pix2tex.models import get_model
from pix2tex.dataset.transforms import test_transform

from .consts import LATEX_CONFIG_FP, MODEL_VERSION, AVAILABLE_MODELS, DOWNLOAD_SOURCE
from .utils import data_dir


logger = logging.getLogger(__name__)


def download_checkpoints(model_info, resizer_model_info, args):
    # adapted from pix2tex.model.checkpoints.get_latest_checkpoint
    ckpt_list = [args.mfr_checkpoint, args.resizer_checkpoint]
    for info, fp in zip([model_info, resizer_model_info], ckpt_list):
        if not os.path.exists(fp):
            logger.warning(f'model file {fp} not found, downloading...')
            info['filename'] = info['url']
            get_model_file(
                info, args.model_root_dir, download_source=DOWNLOAD_SOURCE
            )  # download the .zip file and unzip


def minmax_size(
    img: Image,
    max_dimensions: Tuple[int, int] = None,
    min_dimensions: Tuple[int, int] = None,
) -> Image:
    """Resize or pad an image to fit into given dimensions

    Args:
        img (Image): Image to scale up/down.
        max_dimensions (Tuple[int, int], optional): Maximum dimensions. Defaults to None.
        min_dimensions (Tuple[int, int], optional): Minimum dimensions. Defaults to None.

    Returns:
        Image: Image with correct dimensionality
    """
    if max_dimensions is not None:
        ratios = [a / b for a, b in zip(img.size, max_dimensions)]
        if any([r > 1 for r in ratios]):
            size = np.array(img.size) // max(ratios)
            img = img.resize(size.astype(int), Image.BILINEAR)
    if min_dimensions is not None:
        # hypothesis: there is a dim in img smaller than min_dimensions, and return a proper dim >= min_dimensions
        padded_size = [
            max(img_dim, min_dim) for img_dim, min_dim in zip(img.size, min_dimensions)
        ]
        if padded_size != list(img.size):  # assert hypothesis
            padded_im = Image.new('L', padded_size, 255)
            padded_im.paste(img, img.getbbox())
            img = padded_im
    return img


def find_all_left_or_right(latex, left_or_right='left'):
    left_bracket_infos = []
    prefix_len = len(left_or_right) + 1
    # 匹配出latex中所有的 '\left' 后面跟着的第一个非空格字符，定位它们所在的位置
    for m in re.finditer(rf'\\{left_or_right}\s*\S', latex):
        start, end = m.span()
        # 如果最后一个字符为 "\"，则往前继续匹配，直到匹配到一个非字母的字符
        # 如 "\left \big("
        while latex[end - 1] in ('\\', ' '):
            end += 1
            while end < len(latex) and latex[end].isalpha():
                end += 1
        ori_str = latex[start + prefix_len : end].strip()
        # FIXME: ori_str中可能出现多个 '\left'，此时需要分隔开

        left_bracket_infos.append({'str': ori_str, 'start': start, 'end': end})
        left_bracket_infos.sort(key=lambda x: x['start'])
    return left_bracket_infos


def match_left_right(left_str, right_str):
    """匹配左右括号，如匹配 `\left(` 和 `\right)`。"""
    left_str = left_str.strip().replace(' ', '')[len('left') + 1 :]
    right_str = right_str.strip().replace(' ', '')[len('right') + 1 :]
    # 去掉开头的相同部分
    while left_str and right_str and left_str[0] == right_str[0]:
        left_str = left_str[1:]
        right_str = right_str[1:]

    match_pairs = [
        ('', ''),
        ('(', ')'),
        ('\{', '.'),  # 大括号那种
        ('⟮', '⟯'),
        ('[', ']'),
        ('⟨', '⟩'),
        ('{', '}'),
        ('⌈', '⌉'),
        ('┌', '┐'),
        ('⌊', '⌋'),
        ('└', '┘'),
        ('⎰', '⎱'),
        ('lt', 'gt'),
        ('lang', 'rang'),
        (r'langle', r'rangle'),
        (r'lbrace', r'rbrace'),
        ('lBrace', 'rBrace'),
        (r'lbracket', r'rbracket'),
        (r'lceil', r'rceil'),
        ('lcorner', 'rcorner'),
        (r'lfloor', r'rfloor'),
        (r'lgroup', r'rgroup'),
        (r'lmoustache', r'rmoustache'),
        (r'lparen', r'rparen'),
        (r'lvert', r'rvert'),
        (r'lVert', r'rVert'),
    ]
    return (left_str, right_str) in match_pairs


def post_post_process_latex(latex: str) -> str:
    """对识别结果做进一步处理和修正。"""
    # 把latex中的中文括号全部替换成英文括号
    latex = latex.replace('（', '(').replace('）', ')')
    # 把latex中的中文逗号全部替换成英文逗号
    latex = latex.replace('，', ',')

    left_bracket_infos = find_all_left_or_right(latex, left_or_right='left')
    right_bracket_infos = find_all_left_or_right(latex, left_or_right='right')
    # left 和 right 找配对，left找位置比它靠前且最靠近他的right配对
    for left_bracket_info in left_bracket_infos:
        for right_bracket_info in right_bracket_infos:
            if (
                not right_bracket_info.get('matched', False)
                and right_bracket_info['start'] > left_bracket_info['start']
                and match_left_right(
                    right_bracket_info['str'], left_bracket_info['str']
                )
            ):
                left_bracket_info['matched'] = True
                right_bracket_info['matched'] = True
                break

    for left_bracket_info in left_bracket_infos:
        # 把没有匹配的 '\left'替换为等长度的空格
        left_len = len('left') + 1
        if not left_bracket_info.get('matched', False):
            start_idx = left_bracket_info['start']
            end_idx = start_idx + left_len
            latex = (
                latex[: left_bracket_info['start']]
                + ' ' * (end_idx - start_idx)
                + latex[end_idx:]
            )
    for right_bracket_info in right_bracket_infos:
        # 把没有匹配的 '\right'替换为等长度的空格
        right_len = len('right') + 1
        if not right_bracket_info.get('matched', False):
            start_idx = right_bracket_info['start']
            end_idx = start_idx + right_len
            latex = (
                latex[: right_bracket_info['start']]
                + ' ' * (end_idx - start_idx)
                + latex[end_idx:]
            )

    # 把 latex 中的连续空格替换为一个空格
    latex = re.sub(r'\s+', ' ', latex)
    return latex


class LatexOCR(object):
    """Get a prediction of an image in the easiest way"""

    image_resizer = None
    last_pic = None

    @in_model_path()
    def __init__(
        self,
        *,
        model_name: str = 'mfr',
        model_backend: str = 'pytorch',
        context: str = 'cpu',  # ['cpu', 'gpu', 'cuda']
        model_fp: Optional[Union[str, Path]] = None,
        root: Union[str, Path] = data_dir(),
        arguments: Optional[Dict[str, Any]] = None,
    ):
        """Initialize a LatexOCR model

        Args:
            model_name (str, optional): The name of the model. Defaults to 'mfr'.
            model_backend (str, optional): The model backend. Defaults to 'pytorch'.
            context (str, optional): What device to use for computation, supports `['cpu', 'cuda', 'gpu']`; defaults to `cpu`.
            model_fp (Optional[Union[str, Path]], optional): The model file path. Defaults to None.
            root (Union[str, Path], optional): The model root directory. Defaults to data_dir().
            arguments (Union[Namespace, Munch], optional): Special model parameters. Defaults to None.
        """
        model_root_dir = Path(root) / MODEL_VERSION / 'formula'
        model_info = AVAILABLE_MODELS.get_info(model_name, model_backend)
        resizer_model_info = AVAILABLE_MODELS.get_info('resizer', 'pytorch')

        def_arguments = {
            'config': LATEX_CONFIG_FP,
            'mfr_checkpoint': model_root_dir / model_info['fn'],
            'resizer_checkpoint': model_root_dir / resizer_model_info['fn'],
            'model_root_dir': model_root_dir,
            # 'no_cuda': True,
            'no_resize': False,
            'device': context,
        }
        if arguments is not None:
            if 'model_fp' in arguments:
                arguments['mfr_checkpoint'] = arguments.pop('model_fp')
            def_arguments.update(arguments)
        if model_fp is not None:
            def_arguments['mfr_checkpoint'] = model_fp
        arguments = def_arguments

        arguments = Munch(arguments)
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
        with open(arguments.config, 'r') as f:
            params = yaml.load(f, Loader=yaml.FullLoader)
        self.args = parse_args(Munch(params))
        self.args.update(**vars(arguments))
        self.args.wandb = False
        # self.args.device = (
        #     'cuda' if torch.cuda.is_available() and not self.args.no_cuda else 'cpu'
        # )
        download_checkpoints(model_info, resizer_model_info, self.args)

        self.model = get_model(self.args)
        self.model.load_state_dict(
            torch.load(self.args.mfr_checkpoint, map_location=self.args.device)
        )
        logger.info(f'use model: {self.args.mfr_checkpoint}')
        self.model.eval()

        if not self.args.no_resize and os.path.isfile(self.args.resizer_checkpoint):
            self.image_resizer = ResNetV2(
                layers=[2, 3, 3],
                num_classes=max(self.args.max_dimensions) // 32,
                global_pool='avg',
                in_chans=1,
                drop_rate=0.05,
                preact=True,
                stem_type='same',
                conv_layer=StdConv2dSame,
            ).to(self.args.device)
            self.image_resizer.load_state_dict(
                torch.load(self.args.resizer_checkpoint, map_location=self.args.device)
            )
            logger.info(f'use model: {self.args.resizer_checkpoint}')
            self.image_resizer.eval()
        self.tokenizer = PreTrainedTokenizerFast(tokenizer_file=self.args.tokenizer)

    @in_model_path()
    def __call__(self, img=None, resize=True) -> str:
        """Get a prediction from an image

        Args:
            img (Image, optional): Image to predict. Defaults to None.
            resize (bool, optional): Whether to call the resize model. Defaults to True.

        Returns:
            str: predicted Latex code
        """
        if type(img) is bool:
            img = None
        if img is None:
            if self.last_pic is None:
                print('Provide an image.')
                return ''
            else:
                img = self.last_pic.copy()
        else:
            self.last_pic = img.copy()
        img = minmax_size(pad(img), self.args.max_dimensions, self.args.min_dimensions)
        if (self.image_resizer is not None and not self.args.no_resize) and resize:
            with torch.no_grad():
                input_image = img.convert('RGB').copy()
                r, w, h = 1, input_image.size[0], input_image.size[1]
                for _ in range(10):
                    h = int(h * r)  # height to resize
                    try:
                        img = pad(
                            minmax_size(
                                input_image.resize(
                                    (w, h),
                                    Image.Resampling.BILINEAR
                                    if r > 1
                                    else Image.Resampling.LANCZOS,
                                ),
                                self.args.max_dimensions,
                                self.args.min_dimensions,
                            )
                        )
                        t = test_transform(image=np.array(img.convert('RGB')))['image'][
                            :1
                        ].unsqueeze(0)
                        w = (
                            self.image_resizer(t.to(self.args.device)).argmax(-1).item()
                            + 1
                        ) * 32
                        logger.debug((r, img.size, (w, int(input_image.size[1] * r))))
                    except Exception as e:
                        logger.warning(e)
                        break

                    if w == img.size[0]:
                        break
                    r = w / img.size[0]
        else:
            img = np.array(pad(img).convert('RGB'))
            t = test_transform(image=img)['image'][:1].unsqueeze(0)
        im = t.to(self.args.device)

        dec = self.model.generate(
            im.to(self.args.device), temperature=self.args.get('temperature', 0.25)
        )
        pred = post_process(token2str(dec, self.tokenizer)[0])
        pred = post_post_process_latex(pred)
        return pred



import json
import os
import sys

import cv2
import numpy as np
from shapely.geometry import Polygon
from tabulate import tabulate
import time

def visual_badcase(image_name, pred_list, label_list, output_dir="visual_badcase", info=None, prefix=''):
    """
    """
    image_name = image_name + '.jpg'
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    image_dir = os.path.abspath('../../data/huntie/test_images/')
    image_path = os.path.join(image_dir, image_name)
    img = cv2.imread(image_path)
    if img is None:
        print("--> Warning: skip, given image dir NOT exists: {}".format(image_path))
        return None

    font = cv2.FONT_HERSHEY_SIMPLEX
    img = cv2.imread(image_path)
    for label in label_list:
        points, class_id = label[:8], label[8]
        pts = np.array(points).reshape((1, -1, 2)).astype(np.int32)
        cv2.polylines(img, pts, isClosed=True, color=(0, 255, 0), thickness=3)
        cv2.putText(img, "gt:" + str(class_id), tuple(pts[0][0].tolist()), font, 1, (0, 255, 0), 2)

    for label in pred_list:
        points, class_id = label[:8], label[8]
        pts = np.array(points).reshape((1, -1, 2)).astype(np.int32)
        cv2.polylines(img, pts, isClosed=True, color=(255, 0, 0), thickness=3)
        cv2.putText(img, "pred:" + str(class_id), tuple(pts[0][-1].tolist()), font, 1, (255, 0, 0), 2)

    if info is not None:
        cv2.putText(img, str(info), (40, 40), font, 1, (0, 0, 255), 2)
    output_path = os.path.join(output_dir, prefix + os.path.basename(image_path))
    print("--> info: visualizing badcase: {}".format(output_path))
    cv2.imwrite(output_path, img)


def load_gt_from_json(json_path):
    """
    """
    with open(json_path) as f:
        gt_info = json.load(f)
    gt_image_list = gt_info["images"]
    gt_anno_list = gt_info["annotations"]

    id_to_image_info = {}
    for image_item in gt_image_list:
        id_to_image_info[image_item['id']] = {
            "file_name": image_item['file_name'],
            "group_name": image_item.get("group_name", "huntie")
        }

    group_info = {}
    for annotation_item in gt_anno_list:
        image_info = id_to_image_info[annotation_item['image_id']]
        image_name, group_name = image_info["file_name"], image_info["group_name"]

        if group_name not in group_info:
            group_info[group_name] = {}
        if image_name not in group_info[group_name]:
            group_info[group_name][image_name] = []
        anno_info = {
            "category_id": annotation_item["category_id"],
            "poly": annotation_item["poly"],
            "secondary_id": annotation_item.get("secondary_id", -1),
            "direction_id": annotation_item.get("direction_id", -1)
        }
        group_info[group_name][image_name].append(anno_info)

    group_info_str = ", ".join(["{}[{}]".format(k, len(v)) for k, v in group_info.items()])
    print("--> load {} groups: {}".format(len(group_info.keys()), group_info_str))
    return group_info

def save_res_to_file(table_head, table_body_sorted):
    with open('val_out.txt', 'a') as fout:
        fout.write(time.strftime('%Y-%m-%d-%H-%M') + '\n')
        fout.write('\t'.join(table_head) + '\n')
        for line in table_body_sorted:
            new_line = []
            for ele in line:
                if isinstance(ele, int):
                    new_line.append('{:d}'.format(ele))
                elif isinstance(ele, float):
                    new_line.append('{:.6f}'.format(ele))
                elif isinstance(ele, str):
                    new_line.append(ele)
            fout.write('\t'.join(new_line) + '\n')

def calc_iou(label, detect):
    label_box = []
    detect_box = []

    d_area = []
    for i in range(0, len(detect)):
        pred_poly = detect[i]["poly"]
        box_det = []
        for k in range(0, 4):
            box_det.append([pred_poly[2 * k], pred_poly[2 * k + 1]])
        detect_box.append(box_det)
        try:
            poly = Polygon(box_det)
            d_area.append(poly.area)
        except:
            print('invalid detects', pred_poly)
            exit(-1)

    l_area = []
    for i in range(0, len(label)):
        gt_poly = label[i]["poly"]
        box_gt = []
        for k in range(4):
            box_gt.append([gt_poly[2 * k], gt_poly[2 * k + 1]])
        label_box.append(box_gt)
        try:
            poly = Polygon(box_gt)
            l_area.append(poly.area)
        except:
            print('invalid detects', gt_poly)
            exit(-1)

    ol_areas = []
    for i in range(0, len(detect_box)):
        ol_areas.append([])
        poly1 = Polygon(detect_box[i])
        for j in range(0, len(label_box)):
            poly2 = Polygon(label_box[j])
            try:
                ol_area = poly2.intersection(poly1).area
            except:
                print('invaild pair', detect_box[i], label_box[j])
                ol_areas[i].append(0.0)
            else:
                ol_areas[i].append(ol_area)

    d_ious = [0.0] * len(detect_box)
    l_ious = [0.0] * len(label_box)
    det2label_idx = [-1] * len(detect_box) # 每个检测框iou最大标注框的index
    for i in range(0, len(detect_box)):
        for j in range(0, len(label_box)):
            if int(label[j]["category_id"]) == int(detect[i]["category_id"]):
                # iou = min(ol_areas[i][j] / (d_area[i] + 1e-10), ol_areas[i][j] / (l_area[j] + 1e-10))
                iou = ol_areas[i][j] / (d_area[i] + l_area[j] - ol_areas[i][j] + 1e-10)
            else:
                iou = 0
            det2label_idx[i] = j if iou > d_ious[i] else det2label_idx[i]
            d_ious[i] = max(d_ious[i], iou)
            l_ious[j] = max(l_ious[j], iou)
    return l_ious, d_ious, det2label_idx


def eval(instance_info):
    img_name, label_info = instance_info
    label = label_info['gt']
    detect = label_info['det']
    l_ious, d_ious, det2label_idx = calc_iou(label, detect)
    return [img_name, d_ious, l_ious, detect, label, det2label_idx]


def static_with_class(rets, iou_thresh=0.7, is_verbose=True, map_info=None):
    if is_verbose:
        table_head = ['Class_id', 'Class_name', 'Pre_hit', 'Pre_num', 'GT_hit', 'GT_num', 'Precision', 'Recall', 'F-score', 'All_recalled', 'Img_num', 'Acc.']
    else:
        table_head = ['Class_id', 'Class_name', 'Precision', 'Recall', 'F-score']
    table_body = []
    class_dict = {}
    all_dict = {} # 用以统计合计结果
    all_dict['dm'] = 0
    all_dict['dv'] = 0
    all_dict['lm'] = 0
    all_dict['lv'] = 0
    all_dict['Img_num'] = 0
    all_dict['All_recalled'] = 0

    no_need_keys = ['group_name', 'poly', 'score' , 'category_id']
    # import pdb; pdb.set_trace()
    extra_keys = [_ for _ in rets[0][4][0].keys() if _ not in no_need_keys]
    # extra_table_heads = [[_] for _ in rets[0][4][0].keys() if _ not in no_need_keys]
    extra_table_heads = {}
    extra_dict = {}
    extra_table_body = {}
    for key in extra_keys:
        extra_table_heads[key] = [key, 'Name', 'Pre_hit', 'Pre_num', 'GT_hit', 'GT_num', 'Precision', 'Recall', 'F-score']
        extra_dict[key] = {}
        extra_table_body[key] = []
        # _ += ['Pre_hit', 'Pre_num', 'GT_hit', 'GT_num', 'Precision', 'Recall', 'F-score']

    # pdb.set_trace()
    for i in range(len(rets)):
        img_name, d_ious, l_ious, detects, labels, det2label_idx = rets[i]
        item_lv, item_dv, item_dm, item_lm = 0, 0, 0, 0            
        current_dict = {}

        for label in labels:
            item_lv += 1
            category_id = label["category_id"]
            if category_id not in class_dict:
                class_dict[category_id] = {}
                class_dict[category_id]['dm'] = 0
                class_dict[category_id]['dv'] = 0
                class_dict[category_id]['lm'] = 0
                class_dict[category_id]['lv'] = 0
                class_dict[category_id]['Img_num'] = 0
                class_dict[category_id]['All_recalled'] = 0
            class_dict[category_id]['lv'] += 1

        category_container = []
        for label in labels:
            if label['category_id'] not in category_container:
                category_container.append(label['category_id'])
        for category_id in category_container:
            class_dict[category_id]['Img_num'] += 1
            current_dict[category_id] = {'dm':0, 'dv':0, 'lm':0, 'lv':0, 'Img_num':0, 'All_recalled':0}
        # 统计各额外key的id list和label、detect中检出的量
        for key in extra_keys:
            for label in labels:
                if label[key] not in extra_dict[key] and label[key] != -1:
                    extra_dict[key][label[key]] = {'dm':0, 'dv':0, 'lm':0, 'lv':0}
            for det in detects:
                if det[key] not in extra_dict[key] and det[key] != -1:
                    extra_dict[key][det[key]] = {'dm':0, 'dv':0, 'lm':0, 'lv':0}
            for label in labels:
                if label[key] != -1:
                    extra_dict[key][label[key]]['lv'] += 1
            for det in detects:
                if det[key] != -1:
                    try:
                        extra_dict[key][det[key]]['dv'] += 1
                    except:
                        import pdb; pdb.set_trace()

        for label in labels:
            current_dict[label['category_id']]['lv'] += 1
        for det in detects:
            current_dict[label['category_id']]['dv'] += 1

        for det in detects:
            item_dv += 1
            category_id = det["category_id"]
            if category_id not in class_dict:
                print("--> category_id not exists in gt: {}".format(category_id))
                continue
            class_dict[category_id]['dv'] += 1

        for idx, iou in enumerate(d_ious):
            if iou >= iou_thresh:
                item_dm += 1
                class_dict[detects[idx]["category_id"]]['dm'] += 1
                current_dict[detects[idx]["category_id"]]['dm'] += 1

                for key in extra_keys:
                    if labels[det2label_idx[idx]][key] != -1 and detects[idx][key] == labels[det2label_idx[idx]][key]:
                        extra_dict[key][detects[idx][key]]['dm'] += 1
                        extra_dict[key][detects[idx][key]]['lm'] += 1
                        
        for idx, iou in enumerate(l_ious):
            if iou >= iou_thresh:
                item_lm += 1
                class_dict[labels[idx]["category_id"]]['lm'] += 1
                current_dict[labels[idx]["category_id"]]['lm'] += 1


        
        # 将recall append到结果list当中
        item_r = item_lm / (item_lv + 1e-6)
        # item_p = item_dm / (item_dv + 1e-6)
        # item_f = 2 * item_p * item_r / (item_p + item_r + 1e-6)
        # if (1 - item_r) < 1e-5:
        #     class_dict[category_id]['All_recalled'] += 1
        rets[i].append(item_r)

        # 计算各个box类别全召回率
        for category_id in category_container:
            id_recall = current_dict[category_id]['lm'] / (current_dict[category_id]['lv'] + 1e-6)
            if (1 - id_recall) < 1e-5:
                class_dict[category_id]['All_recalled'] += 1

        # 计算所有类别总计的全召回率
        # import pdb; pdb.set_trace()
        all_dict['dv'] += item_dv
        all_dict['lv'] += item_lv
        all_dict['dm'] += item_dm
        all_dict['lm'] += item_lm
        all_dict['Img_num'] += 1
        if (1 - item_r) < 1e-5:
            all_dict['All_recalled'] += 1
        
        # if img_name == 'train10w_val2w_69008cd9828a455fb1bf751a95ad8921.jpg':
        #     import pdb; pdb.set_trace()
        # if item_r
        # if item_f < 0.97 and is_save_badcase:
        #     prefix = '_'.join(map(str, sorted(list(badcase_class_set)))) + '_'
        #     item_info = "IOU{}, {}, {}, {}".format(iou_thresh, item_r, item_p, item_f)
        #     visual_badcase(img_name, detects, labels, output_dir="visual_badcase", info=item_info, prefix=prefix)

    dm, dv, lm, lv, total, recalled = 0, 0, 0, 0, 0, 0
    map_info = {} if map_info is None else map_info
    for key in class_dict.keys():
        dm += class_dict[key]['dm']
        dv += class_dict[key]['dv']
        lm += class_dict[key]['lm']
        lv += class_dict[key]['lv']
        recalled += class_dict[category_id]['All_recalled']
        total += class_dict[key]['Img_num']
        p = class_dict[key]['dm'] / (class_dict[key]['dv'] + 1e-6)
        r = class_dict[key]['lm'] / (class_dict[key]['lv'] + 1e-6)
        fscore = 2 * p * r / (p + r + 1e-6)
        acc = class_dict[key]['All_recalled'] / (class_dict[key]['Img_num'] + 1e-6)
        if is_verbose:
            table_body.append((key, map_info.get("primary_map", {}).get(str(key), str(key)), class_dict[key]['dm'],
                               class_dict[key]['dv'], class_dict[key]['lm'], class_dict[key]['lv'], p, r, fscore,
                               class_dict[category_id]['All_recalled'], class_dict[key]['Img_num'], acc))
        else:
            table_body.append((key,  map_info.get(str(key), str(key)), p, r, fscore))

    p = dm / (dv + 1e-6)
    r = lm / (lv + 1e-6)
    f = 2 * p * r / (p + r + 1e-6)
    acc = recalled / (total + 1e-6)
    table_body_sorted = sorted(table_body, key=lambda x: int((x[0])))
    if is_verbose:
        table_body_sorted.append(('IOU_{}'.format(iou_thresh), 'average', dm, dv, lm, lv, p, r, f,
                                  all_dict['All_recalled'], all_dict['Img_num'], (all_dict['All_recalled']/all_dict['Img_num']+1e-6)))
    else:
        table_body_sorted.append(('IOU_{}'.format(iou_thresh), 'average', p, r, f))
    # import pdb; pdb.set_trace()
    save_res_to_file(table_head, table_body_sorted)
    print(tabulate(table_body_sorted, headers=table_head, tablefmt='pipe'))
    # ---------------print(extra_keys)
    for _key in extra_dict.keys():
        dm, dv, lm, lv = 0, 0, 0, 0
        for key in extra_dict[_key].keys():
            dm += extra_dict[_key][key]['dm']
            dv += extra_dict[_key][key]['dv']
            lm += extra_dict[_key][key]['lm']
            lv += extra_dict[_key][key]['lv']
            # 找当前key对应的map_info key的name
            map_name = ''
            for candidate_name in map_info.keys():
                if candidate_name.split('_')[0] == _key.split('_')[0]:
                    map_name = candidate_name

            precision = extra_dict[_key][key]['dm'] / (extra_dict[_key][key]['dv'] + 1e-6)
            recall = extra_dict[_key][key]['lm'] / (extra_dict[_key][key]['lv'] + 1e-6)
            fscore = 2 * precision * recall / (precision + recall + 1e-6)
            if map_name == '': # 没有在map_info中找到对应类表
                extra_table_body[_key].append((key, '', extra_dict[_key][key]['dm'], extra_dict[_key][key]['dv'],
                                                extra_dict[_key][key]['lm'], extra_dict[_key][key]['lv'],
                                            precision, recall, fscore))
            else:
                extra_table_body[_key].append((key, map_info.get(map_name, {}).get(str(key), str(key)), extra_dict[_key][key]['dm'], extra_dict[_key][key]['dv'],
                                                extra_dict[_key][key]['lm'], extra_dict[_key][key]['lv'],
                                            precision, recall, fscore))                
        extra_table_body[_key] = sorted(extra_table_body[_key], key=lambda x: int((x[0])))
        p = dm / (dv + 1e-6)
        r = lm / (lv + 1e-6)
        f = 2 * p * r / (p + r + 1e-6)
        extra_table_body[_key].append((key, 'average', dm, dv, lm, lv, p, r, f))
    # import pdb; pdb.set_trace()
    for _key in extra_keys:
        save_res_to_file(extra_table_heads[_key], extra_table_body[_key])
        print(tabulate(extra_table_body[_key], headers=extra_table_heads[_key], tablefmt='pipe'))

    return [table_head] + table_body_sorted


def multiproc(func, task_list, proc_num=30, retv=True, progress_bar=False):
    from multiprocessing import Pool
    pool = Pool(proc_num)

    rets = []
    if progress_bar:
        import tqdm
        with tqdm.tqdm(total=len(task_list)) as t:
            for ret in pool.imap(func, task_list):
                rets.append(ret)
                t.update(1)
    else:
        for ret in pool.imap(func, task_list):
            rets.append(ret)

    pool.close()
    pool.join()

    if retv:
        return rets


def eval_and_show(label_dict, detect_dict, output_dir, iou_thresh=0.7, map_info=None):
    """
    """
    evaluation_group_info = {}
    for group_name, gt_info in label_dict.items():
        group_pair_list = []
        for file_name, value_list in gt_info.items():
            if file_name not in detect_dict:
                # print("--> missing pred:", file_name)
                continue
            group_pair_list.append([file_name, {'gt': gt_info[file_name], 'det': detect_dict[file_name]}])
        evaluation_group_info[group_name] = group_pair_list

    res_info_all = {}
    for group_name, group_pair_list in evaluation_group_info.items():
        print(" ------- group name: {} -----------".format(group_name))
        rets = multiproc(eval, group_pair_list, proc_num=16)
        # import pdb; pdb.set_trace()
        group_name_map_info = map_info.get(group_name, None) if map_info is not None else None
        res_info = static_with_class(rets, iou_thresh=iou_thresh, map_info=group_name_map_info)
        res_info_all[group_name] = res_info

    evaluation_res_info_path = os.path.join(output_dir, "results_val.json")
    with open(evaluation_res_info_path, "w") as f:
        json.dump(res_info_all, f, ensure_ascii=False, indent=4)
    print("--> info: evaluation result is saved at {}".format(evaluation_res_info_path))
    return rets

if __name__ == "__main__":

    if len(sys.argv) != 5:
        print("Usage: python {} gt_json_path pred_json_path output_dir  iou_thresh".format(__file__))
        exit(-1)
    else:
        print('--> info: {}'.format(sys.argv))
        gt_json_path, pred_json_path, output_dir, iou_thresh = sys.argv[1], sys.argv[2], sys.argv[3], float(sys.argv[4])

    label_dict = load_gt_from_json(gt_json_path)
    with open(pred_json_path, "r") as f:
        detect_dict = json.load(f)
    res_info = eval_and_show(label_dict, detect_dict, output_dir, iou_thresh=iou_thresh, map_info=None)




# coding: utf-8
import time
import numpy as np
import torch

# from external.nms import soft_nms
from ..external.shapelyNMS import pnms
from ..models.decode import ctdet_4ps_decode, ctdet_cls_decode
from ..models.utils import flip_tensor
from ..utils.post_process import ctdet_4ps_post_process
from .base_detector_subfield import BaseDetector


class CtdetDetector_Subfield(BaseDetector):
    def __init__(self, opt):
        super(CtdetDetector_Subfield, self).__init__(opt)

    def process(self, images, return_time=False):
        # import ipdb;ipdb.set_trace()
        with torch.no_grad():
            output = self.model(images)[-1]
            if self.opt.convert_onnx == 1:
                # torch.cuda.synchronize()
                inputs = ['data']
                outputs = [
                    'hm.0.sigmoid',
                    'hm.0.maxpool',
                    'cls.0.sigmoid',
                    'ftype.0.sigmoid',
                    'wh.2',
                    'reg.2',
                    'hm_sub.0.sigmoid',
                    'hm_sub.0.maxpool',
                    'wh_sub.2',
                    'reg_sub.2',
                ]
                dynamic_axes = {
                    'data': {2: 'h', 3: 'w'},
                    'hm.0.sigmoid': {2: 'H', 3: 'W'},
                    'hm.0.maxpool': {2: 'H', 3: 'W'},
                    'cls.0.sigmoid': {2: 'H', 3: 'W'},
                    'ftype.0.sigmoid': {2: 'H', 3: 'W'},
                    'wh.2': {2: 'H', 3: 'W'},
                    'reg.2': {2: 'H', 3: 'W'},
                    'hm_sub.0.sigmoid': {2: 'H', 3: 'W'},
                    'hm_sub.0.maxpool': {2: 'H', 3: 'W'},
                    'wh_sub.2': {2: 'H', 3: 'W'},
                    'reg_sub.2': {2: 'H', 3: 'W'},
                }

                onnx_path = self.opt.onnx_path
                if self.opt.onnx_path == "auto":
                    onnx_path = "{}_{}cls_{}ftype.onnx".format(
                        self.opt.dataset,
                        self.opt.num_classes,
                        self.opt.num_secondary_classes,
                    )

                torch.onnx.export(
                    self.model,
                    images,
                    onnx_path,
                    input_names=inputs,
                    output_names=outputs,
                    dynamic_axes=dynamic_axes,
                    do_constant_folding=True,
                    opset_version=10,
                )
                print("--> info: onnx is saved at: {}".format(onnx_path))
                cls = output['cls_sigmoid']
                hm = output['hm_sigmoid']
                ftype = output['ftype_sigmoid']

                # add sub
                hm_sub = output['hm_sigmoid_sub']
            else:
                hm = output['hm'].sigmoid_()
                cls = output['cls'].sigmoid_()
                ftype = output['ftype'].sigmoid_()

                # add sub
                hm_sub = output['hm_sub'].sigmoid_()

            wh = output['wh']
            reg = output['reg'] if self.opt.reg_offset else None

            # add sub
            wh_sub = output['wh_sub']
            reg_sub = output['reg_sub'] if self.opt.reg_offset else None

            if self.opt.flip_test:
                hm = (hm[0:1] + flip_tensor(hm[1:2])) / 2
                wh = (wh[0:1] + flip_tensor(wh[1:2])) / 2
                reg = reg[0:1] if reg is not None else None
            # torch.cuda.synchronize()
            forward_time = time.time()
            # return dets [bboxes, scores, clses]
            # breakpoint()
            dets, inds = ctdet_4ps_decode(hm, wh, reg=reg, K=self.opt.K)

            # add sub
            dets_sub, inds_sub = ctdet_4ps_decode(
                hm_sub, wh_sub, reg=reg_sub, K=self.opt.K
            )

            box_cls = ctdet_cls_decode(cls, inds)
            box_ftype = ctdet_cls_decode(ftype, inds)
            clses = torch.argmax(box_cls, dim=2, keepdim=True)
            ftypes = torch.argmax(box_ftype, dim=2, keepdim=True)
            dets = np.concatenate(
                (
                    dets.detach().cpu().numpy(),
                    clses.detach().cpu().numpy(),
                    ftypes.detach().cpu().numpy(),
                ),
                axis=2,
            )
            dets = np.array(dets)

            # add subfield
            dets_sub = np.concatenate(
                (
                    dets_sub.detach().cpu().numpy(),
                    clses.detach().cpu().numpy(),
                    ftypes.detach().cpu().numpy(),
                ),
                axis=2,
            )
            dets_sub = np.array(dets_sub)
            dets_sub[:, :, -3] += 11

            corner = 0

        if return_time:
            return output, dets, dets_sub, corner, forward_time
        else:
            return output, dets, dets_sub

    def post_process(self, dets, corner, meta, scale=1):
        if self.opt.nms:
            detn = pnms(dets[0], self.opt.scores_thresh)
            if detn.shape[0] > 0:
                dets = detn.reshape(1, -1, detn.shape[1])
        k = dets.shape[2] if dets.shape[1] != 0 else 0
        if dets.shape[1] != 0:
            dets = dets.reshape(1, -1, dets.shape[2])
            # return dets is list and what in dets is dict. key of dict is classes, value of dict is [bbox,score]
            dets = ctdet_4ps_post_process(
                dets.copy(),
                [meta['c']],
                [meta['s']],
                meta['out_height'],
                meta['out_width'],
                self.opt.num_classes,
            )
            for j in range(1, self.num_classes + 1):
                dets[0][j] = np.array(dets[0][j], dtype=np.float32).reshape(-1, k)
                dets[0][j][:, :8] /= scale
        else:
            ret = {}
            dets = []
            for j in range(1, self.num_classes + 1):
                ret[j] = np.array([0] * k, dtype=np.float32)  # .reshape(-1, k)
            dets.append(ret)
        return dets[0], corner

    def merge_outputs(self, detections):
        results = {}
        for j in range(1, self.num_classes + 1):
            results[j] = np.concatenate(
                [detection[j] for detection in detections], axis=0
            ).astype(np.float32)
            # if len(self.scales) > 1 or self.opt.nms:
            #  results[j] = pnms(results[j],self.opt.nms_thresh)
        shape_num = 0
        for j in range(1, self.num_classes + 1):
            shape_num = shape_num + len(results[j])
        if shape_num != 0:
            # print(np.array(results[1]))
            scores = np.hstack(
                [results[j][:, 8] for j in range(1, self.num_classes + 1)]
            )
        else:
            scores = []
        if len(scores) > self.max_per_image:
            kth = len(scores) - self.max_per_image
            thresh = np.partition(scores, kth)[kth]
            for j in range(1, self.num_classes + 1):
                keep_inds = results[j][:, 8] >= thresh
                results[j] = results[j][keep_inds]
        return results

    def debug(self, debugger, images, dets, output, scale=1):
        # detection = dets.detach().cpu().numpy().copy()
        detection = dets.copy()
        detection[:, :, :8] *= self.opt.down_ratio
        for i in range(1):
            img = images[i].detach().cpu().numpy().transpose(1, 2, 0)
            img = ((img * self.std + self.mean) * 255).astype(np.uint8)
            pred = debugger.gen_colormap(output['hm'][i].detach().cpu().numpy())
            debugger.add_blend_img(img, pred, 'pred_hm_{:.1f}'.format(scale))
            debugger.add_img(img, img_id='out_pred_{:.1f}'.format(scale))
            # pdb.set_trace()
            for k in range(len(dets[i])):
                if detection[i, k, 8] > self.opt.center_thresh:
                    debugger.add_4ps_coco_bbox(
                        detection[i, k, :8],
                        detection[i, k, -1],
                        detection[i, k, 8],
                        img_id='out_pred_{:.1f}'.format(scale),
                    )

    def show_results(self, debugger, image, results, Corners, image_name):
        debugger.add_img(image, img_id='ctdet')
        count = 0
        for j in range(1, self.num_classes + 1):
            for bbox in results[j]:
                if bbox[8] > self.opt.scores_thresh:
                    count += 1
                    # print("bbox info:",j-1,  bbox.tolist())
                    # print(j-1)
                    debugger.add_4ps_coco_bbox(
                        bbox, j - 1, bbox[8], show_txt=True, img_id='ctdet'
                    )
        debugger.save_all_imgs(image_name, './outputs/')



# coding: utf-8
#! pip install pillow transformers optimum[onnxruntime]
from PIL import Image
from transformers import TrOCRProcessor
from optimum.onnxruntime import ORTModelForVision2Seq

processor = TrOCRProcessor.from_pretrained('breezedeus/pix2text-mfr')
model = ORTModelForVision2Seq.from_pretrained('breezedeus/pix2text-mfr', use_cache=False)

image_fps = [
    # 'https://github.com/breezedeus/Pix2Text/blob/main/docs/examples/formula.jpg?raw=true',
    'docs/examples/formula.jpg',
    # 'examples/0000186.png',
]
images = [Image.open(fp).convert('RGB') for fp in image_fps]
pixel_values = processor(images=images, return_tensors="pt").pixel_values
generated_ids = model.generate(pixel_values)
generated_text = processor.batch_decode(generated_ids, skip_special_tokens=True)
print(f'generated_ids: {generated_ids}, \ngenerated text: {generated_text}')




# coding: utf-8

import requests


def main():
    url = 'http://0.0.0.0:8503/pix2text'

    image_fp = 'docs/examples/page2.png'
    # image_fp = 'docs/examples/mixed.jpg'
    # image_fp = 'docs/examples/math-formula-42.png'
    # image_fp = 'docs/examples/english.jpg'
    data = {
        "file_type": "page",
        "resized_shape": 768,
        "embed_sep": " $,$ ",
        "isolated_sep": "$$\n, \n$$"
    }
    files = {
        "image": (image_fp, open(image_fp, 'rb'), 'image/jpeg')
    }

    r = requests.post(url, data=data, files=files)

    outs = r.json()['results']
    out_md_dir = r.json()['output_dir']
    if isinstance(outs, str):
        only_text = outs
    else:
        only_text = '\n'.join([out['text'] for out in outs])
    print(f'{only_text=}')
    print(f'{out_md_dir=}')


if __name__ == '__main__':
    main()



# coding: utf-8
# [Pix2Text](https://github.com/breezedeus/pix2text): an Open-Source Alternative to Mathpix.
# Copyright (C) 2022-2024, [Breezedeus](https://www.breezedeus.com).

import logging
import re
from itertools import chain
from pathlib import Path
from typing import Dict, Any, Optional, Union, List, Sequence
from copy import copy, deepcopy

from PIL import Image
import numpy as np
import torch
from cnstd import LayoutAnalyzer
from cnstd.yolov7.general import box_partial_overlap
from spellchecker import SpellChecker

from .utils import (
    sort_boxes,
    merge_adjacent_bboxes,
    adjust_line_height,
    adjust_line_width,
    rotated_box_to_horizontal,
    is_valid_box,
    list2box,
    select_device,
    prepare_imgs,
    merge_line_texts,
    remove_overlap_text_bbox,
    y_overlap,
)
from .ocr_engine import prepare_ocr_engine, TextOcrEngine
from .latex_ocr import LatexOCR
from .utils import (
    read_img,
    save_layout_img,
)

logger = logging.getLogger(__name__)


DEFAULT_CONFIGS = {
    'mfd': {},
    'text': {},
    'formula': {},
}
# see: https://pypi.org/project/pyspellchecker
CHECKER_SUPPORTED_LANGUAGES = {
    'en',
    'es',
    'fr',
    'pt',
    'de',
    'it',
    'ru',
    'ar',
    'eu',
    'lv',
    'nl',
}


class TextFormulaOCR(object):
    def __init__(
        self,
        *,
        text_ocr: Optional[TextOcrEngine] = None,
        mfd: Optional[LayoutAnalyzer] = None,
        latex_ocr: Optional[LatexOCR] = None,
        spellchecker: Optional[SpellChecker] = None,
        **kwargs,
    ):
        if text_ocr is None:
            text_config = deepcopy(DEFAULT_CONFIGS['text'])
            device = select_device(device=None)
            text_config['context'] = device
            logger.warning(
                f'text_ocr must not be None. Using default text_ocr engine instead, with config: {text_config}.'
            )
            text_ocr = prepare_ocr_engine(
                languages=('en', 'ch_sim'), ocr_engine_config=text_config
            )
        # if mfd is None or latex_ocr is None:
        #     default_ocr = TextFormulaOCR.from_config()
        #     mfd = default_ocr.mfd if mfd is None else mfd
        #     text_ocr = default_ocr.text_ocr if text_ocr is None else text_ocr
        #     latex_ocr = default_ocr.latex_ocr if latex_ocr is None else latex_ocr
        #     del default_ocr

        self.text_ocr = text_ocr
        self.mfd = mfd
        self.latex_ocr = latex_ocr
        self.spellchecker = spellchecker

    @classmethod
    def from_config(
        cls,
        total_configs: Optional[dict] = None,
        enable_formula: bool = True,
        enable_spell_checker: bool = True,
        device: str = None,
        **kwargs,
    ):
        """
        Args:
            total_configs (dict): Configuration information for Pix2Text; defaults to `None`, which means using the default configuration. Usually the following keys are used:

                * languages (str or Sequence[str]): The language code(s) of the text to be recognized; defaults to `('en', 'ch_sim')`.
                * mfd (dict): Configuration information for the Analyzer model; defaults to `None`, which means using the default configuration.
                * text (dict): Configuration information for the Text OCR model; defaults to `None`, which means using the default configuration.
                * formula (dict): Configuration information for Math Formula OCR model; defaults to `None`, which means using the default configuration.
            enable_formula (bool): Whether to enable the capability of Math Formula Detection (MFD) and Recognition (MFR); defaults to True.
            enable_spell_checker (bool): Whether to enable the capability of Spell Checker; defaults to True.
            device (str, optional): What device to use for computation, supports `['cpu', 'cuda', 'gpu', 'mps']`; defaults to None, which selects the device automatically.
            **kwargs (): Reserved for other parameters; not currently used.
        """
        total_configs = total_configs or DEFAULT_CONFIGS
        languages = total_configs.get('languages', ('en', 'ch_sim'))
        text_config = total_configs.get('text', dict())
        mfd_config = total_configs.get('mfd', dict())
        formula_config = total_configs.get('formula', dict())

        device = select_device(device)
        mfd_config, text_config, formula_config = cls.prepare_configs(
            mfd_config, text_config, formula_config, device,
        )

        text_ocr = prepare_ocr_engine(languages, text_config)

        if enable_formula:
            if 'model_name' in mfd_config:
                mfd_config.pop('model_name')
            mfd = LayoutAnalyzer(model_name='mfd', **mfd_config)
            latex_ocr = LatexOCR(**formula_config)
        else:
            mfd = None
            latex_ocr = None

        spellchecker = None
        if enable_spell_checker:
            checker_languages = set(languages) & CHECKER_SUPPORTED_LANGUAGES
            if checker_languages:
                spellchecker = SpellChecker(language=checker_languages)

        return cls(
            text_ocr=text_ocr,
            mfd=mfd,
            latex_ocr=latex_ocr,
            spellchecker=spellchecker,
            **kwargs,
        )

    @classmethod
    def prepare_configs(
        cls, mfd_config, text_config, formula_config, device,
    ):
        def _to_default(_conf, _def_val):
            if not _conf:
                _conf = _def_val
            return _conf

        mfd_config = _to_default(mfd_config, DEFAULT_CONFIGS['mfd'])
        # FIXME
        mfd_config['device'] = device if device != 'mps' else 'cpu'
        text_config = _to_default(text_config, DEFAULT_CONFIGS['text'])
        text_config['context'] = device
        formula_config = _to_default(formula_config, DEFAULT_CONFIGS['formula'])
        formula_config['device'] = device
        return (
            mfd_config,
            text_config,
            formula_config,
        )

    @property
    def languages(self):
        return self.text_ocr.languages

    def __call__(
        self, img: Union[str, Path, Image.Image], **kwargs
    ) -> List[Dict[str, Any]]:
        return self.recognize(img, **kwargs)

    def recognize(
        self, img: Union[str, Path, Image.Image], return_text: bool = True, **kwargs
    ) -> Union[str, List[Dict[str, Any]]]:
        """
        Perform Mathematical Formula Detection (MFD) on the image, and then recognize the information contained in each section.

        Args:
            img (str or Image.Image): an image path, or `Image.Image` loaded by `Image.open()`
            return_text (bool): Whether to return only the recognized text; default value is `True`
            kwargs ():
                * contain_formula (bool): If `True`, the image will be recognized as a mixed image (text and formula). If `False`, it will be recognized as a text; default value is `True`
                * resized_shape (int): Resize the image width to this size for processing; default value is `768`
                * save_analysis_res (str): Save the parsed result image in this file; default value is `None`, which means not to save
                * mfr_batch_size (int): batch size for MFR; When running on GPU, this value is suggested to be set to greater than 1; default value is `1`
                * embed_sep (tuple): Prefix and suffix for embedding latex; only effective when `return_text` is `True`; default value is `(' $', '$ ')`
                * isolated_sep (tuple): Prefix and suffix for isolated latex; only effective when `return_text` is `True`; default value is two-dollar signs
                * line_sep (str): The separator between lines of text; only effective when `return_text` is `True`; default value is a line break
                * auto_line_break (bool): Automatically line break the recognized text; only effective when `return_text` is `True`; default value is `True`
                * det_text_bbox_max_width_expand_ratio (float): Expand the width of the detected text bbox. This value represents the maximum expansion ratio above and below relative to the original bbox height; default value is `0.3`
                * det_text_bbox_max_height_expand_ratio (float): Expand the height of the detected text bbox. This value represents the maximum expansion ratio above and below relative to the original bbox height; default value is `0.2`
                * embed_ratio_threshold (float): The overlap threshold for embed formulas and text lines; default value is `0.6`.
                    When the overlap between an embed formula and a text line is greater than or equal to this threshold,
                    the embed formula and the text line are considered to be on the same line;
                    otherwise, they are considered to be on different lines.
                * formula_rec_kwargs (dict): generation arguments passed to formula recognizer `latex_ocr`; default value is `{}`

        Returns: a str when `return_text` is `True`, or a list of ordered (top to bottom, left to right) dicts when `return_text` is `False`,
            with each dict representing one detected box, containing keys:

                * `type`: The category of the image; Optional: 'text', 'isolated', 'embedding'
                * `text`: The recognized text or Latex formula
                * `score`: The confidence score [0, 1]; the higher, the more confident
                * `position`: Position information of the block, `np.ndarray`, with shape of [4, 2]
                * `line_number`: The line number of the box (first line `line_number==0`), boxes with the same value indicate they are on the same line

        """
        # 对于大图片，把图片宽度resize到此大小；宽度比此小的图片，其实不会放大到此大小，
        # 具体参考：cnstd.yolov7.layout_analyzer.LayoutAnalyzer._preprocess_images 中的 `letterbox` 行
        resized_shape = kwargs.get('resized_shape', 768)
        if isinstance(img, Image.Image):
            img0 = img.convert('RGB')
        else:
            img0 = read_img(img, return_type='Image')
        w, h = img0.size
        ratio = resized_shape / w
        resized_shape = (int(h * ratio), resized_shape)  # (H, W)
        # logger.debug('MFD Result: %s', analyzer_outs)
        analyzer_outs = []
        crop_patches = []
        mf_results = []
        if (
            kwargs.get('contain_formula', True)
            and self.mfd is not None
            and self.latex_ocr is not None
        ):
            analyzer_outs = self.mfd(img0.copy(), resized_shape=resized_shape)
            for mf_box_info in analyzer_outs:
                box = mf_box_info['box']
                xmin, ymin, xmax, ymax = (
                    int(box[0][0]),
                    int(box[0][1]),
                    int(box[2][0]),
                    int(box[2][1]),
                )
                crop_patch = img0.crop((xmin, ymin, xmax, ymax))
                crop_patches.append(crop_patch)

            mfr_batch_size = kwargs.get('mfr_batch_size', 1)
            formula_rec_kwargs = kwargs.get('formula_rec_kwargs', {})
            mf_results = self.latex_ocr.recognize(
                crop_patches, batch_size=mfr_batch_size, **formula_rec_kwargs
            )

        assert len(mf_results) == len(analyzer_outs)

        mf_outs = []
        for mf_box_info, patch_out in zip(analyzer_outs, mf_results):
            text = patch_out['text']
            mf_outs.append(
                {
                    'type': mf_box_info['type'],
                    'text': text,
                    'position': mf_box_info['box'],
                    'score': patch_out['score'],
                }
            )

        masked_img = np.array(img0.copy())
        # 把公式部分mask掉，然后对其他部分进行OCR
        for mf_box_info in analyzer_outs:
            if mf_box_info['type'] in ('isolated', 'embedding'):
                box = mf_box_info['box']
                xmin, ymin = max(0, int(box[0][0]) - 1), max(0, int(box[0][1]) - 1)
                xmax, ymax = (
                    min(img0.size[0], int(box[2][0]) + 1),
                    min(img0.size[1], int(box[2][1]) + 1),
                )
                masked_img[ymin:ymax, xmin:xmax, :] = 255
        masked_img = Image.fromarray(masked_img)

        text_box_infos = self.text_ocr.detect_only(
            np.array(img0), resized_shape=resized_shape
        )
        box_infos = []
        for line_box_info in text_box_infos['detected_texts']:
            # crop_img_info['box'] 可能是一个带角度的矩形框，需要转换成水平的矩形框
            _text_box = rotated_box_to_horizontal(line_box_info['position'])
            if not is_valid_box(_text_box, min_height=8, min_width=2):
                continue
            box_infos.append({'position': _text_box})
        max_width_expand_ratio = kwargs.get('det_text_bbox_max_width_expand_ratio', 0.3)
        if self.text_ocr.name == 'cnocr':
            box_infos: list[dict] = adjust_line_width(
                text_box_infos=box_infos,
                formula_box_infos=mf_outs,
                img_width=img0.size[0],
                max_expand_ratio=max_width_expand_ratio,
            )
        box_infos = remove_overlap_text_bbox(box_infos, mf_outs)

        def _to_iou_box(ori):
            return torch.tensor([ori[0][0], ori[0][1], ori[2][0], ori[2][1]]).unsqueeze(
                0
            )

        embed_ratio_threshold = kwargs.get('embed_ratio_threshold', 0.6)
        total_text_boxes = []
        for line_box_info in box_infos:
            _line_box = _to_iou_box(line_box_info['position'])
            _embed_mfs = []
            for mf_box_info in mf_outs:
                if mf_box_info['type'] == 'embedding':
                    _mf_box = _to_iou_box(mf_box_info['position'])
                    overlap_area_ratio = float(
                        box_partial_overlap(_line_box, _mf_box).squeeze()
                    )
                    if overlap_area_ratio >= embed_ratio_threshold or (
                        overlap_area_ratio > 0
                        and y_overlap(line_box_info, mf_box_info, key='position')
                        > embed_ratio_threshold
                    ):
                        _embed_mfs.append(
                            {
                                'position': _mf_box[0].int().tolist(),
                                'text': mf_box_info['text'],
                                'type': mf_box_info['type'],
                            }
                        )

            ocr_boxes = self._split_line_image(_line_box, _embed_mfs)
            total_text_boxes.extend(ocr_boxes)

        outs = copy(mf_outs)
        for box in total_text_boxes:
            box['position'] = list2box(*box['position'])
            outs.append(box)
        outs = sort_boxes(outs, key='position')
        outs = [merge_adjacent_bboxes(bboxes) for bboxes in outs]
        max_height_expand_ratio = kwargs.get(
            'det_text_bbox_max_height_expand_ratio', 0.2
        )
        outs = adjust_line_height(
            outs, img0.size[1], max_expand_ratio=max_height_expand_ratio
        )

        for line_idx, line_boxes in enumerate(outs):
            for box in line_boxes:
                if box['type'] != 'text':
                    continue
                bbox = box['position']
                xmin, ymin, xmax, ymax = (
                    int(bbox[0][0]),
                    int(bbox[0][1]),
                    int(bbox[2][0]),
                    int(bbox[2][1]),
                )
                crop_patch = np.array(masked_img.crop((xmin, ymin, xmax, ymax)))
                part_res = self.text_ocr.recognize_only(crop_patch)
                box['text'] = part_res['text']
                box['score'] = part_res['score']
            outs[line_idx] = [box for box in line_boxes if box['text'].strip()]

        logger.debug(outs)
        outs = self._post_process(outs)

        outs = list(chain(*outs))
        if kwargs.get('save_analysis_res'):
            save_layout_img(
                img0,
                ('text', 'isolated', 'embedding'),
                outs,
                kwargs.get('save_analysis_res'),
            )

        if return_text:
            embed_sep = kwargs.get('embed_sep', (' $', '$ '))
            isolated_sep = kwargs.get('isolated_sep', ('$$\n', '\n$$'))
            line_sep = kwargs.get('line_sep', '\n')
            auto_line_break = kwargs.get('auto_line_break', True)
            outs = merge_line_texts(
                outs,
                auto_line_break,
                line_sep,
                embed_sep,
                isolated_sep,
                self.spellchecker,
            )

        return outs

    def _post_process(self, outs):
        match_pairs = [
            (',', ',，'),
            ('.', '.。'),
            ('?', '?？'),
        ]
        formula_tag = '^[（\(]\d+(\.\d+)*[）\)]$'

        def _match(a1, a2):
            matched = False
            for b1, b2 in match_pairs:
                if a1 in b1 and a2 in b2:
                    matched = True
                    break
            return matched

        for idx, line_boxes in enumerate(outs):
            if (
                any([_lang in ('ch_sim', 'ch_tra') for _lang in self.languages])
                and len(line_boxes) > 1
                and line_boxes[-1]['type'] == 'text'
                and line_boxes[-2]['type'] != 'text'
            ):
                if line_boxes[-1]['text'].lower() == 'o':
                    line_boxes[-1]['text'] = '。'
            if len(line_boxes) > 1:
                # 去掉边界上多余的标点
                for _idx2, box in enumerate(line_boxes[1:]):
                    if (
                        box['type'] == 'text'
                        and line_boxes[_idx2]['type'] == 'embedding'
                    ):  # if the current box is text and the previous box is embedding
                        if _match(line_boxes[_idx2]['text'][-1], box['text'][0]) and (
                            not line_boxes[_idx2]['text'][:-1].endswith('\\')
                            and not line_boxes[_idx2]['text'][:-1].endswith(r'\end')
                        ):
                            line_boxes[_idx2]['text'] = line_boxes[_idx2]['text'][:-1]
                # 把 公式 tag 合并到公式里面去
                for _idx2, box in enumerate(line_boxes[1:]):
                    if (
                        box['type'] == 'text'
                        and line_boxes[_idx2]['type'] == 'isolated'
                    ):  # if the current box is text and the previous box is embedding
                        if y_overlap(line_boxes[_idx2], box, key='position') > 0.9:
                            if re.match(formula_tag, box['text']):
                                # 去掉开头和结尾的括号
                                tag_text = box['text'][1:-1]
                                line_boxes[_idx2]['text'] = line_boxes[_idx2][
                                    'text'
                                ] + ' \\tag{{{}}}'.format(tag_text)
                                new_xmax = max(
                                    line_boxes[_idx2]['position'][2][0],
                                    box['position'][2][0],
                                )
                                line_boxes[_idx2]['position'][1][0] = line_boxes[_idx2][
                                    'position'
                                ][2][0] = new_xmax
                                box['text'] = ''

            outs[idx] = [box for box in line_boxes if box['text'].strip()]
        return outs

    @classmethod
    def _split_line_image(cls, line_box, embed_mfs):
        # 利用embedding formula所在位置，把单行文字图片切割成多个小段，之后这些小段会分别扔进OCR进行文字识别
        line_box = line_box[0]
        if not embed_mfs:
            return [{'position': line_box.int().tolist(), 'type': 'text'}]
        embed_mfs.sort(key=lambda x: x['position'][0])

        outs = []
        start = int(line_box[0])
        xmax, ymin, ymax = int(line_box[2]), int(line_box[1]), int(line_box[-1])
        for mf in embed_mfs:
            _xmax = min(xmax, int(mf['position'][0]) + 1)
            if start + 8 < _xmax:
                outs.append({'position': [start, ymin, _xmax, ymax], 'type': 'text'})
            start = int(mf['position'][2])
            if _xmax >= xmax:
                break
        if start < xmax:
            outs.append({'position': [start, ymin, xmax, ymax], 'type': 'text'})
        return outs

    def recognize_text(
        self,
        imgs: Union[str, Path, Image.Image, List[str], List[Path], List[Image.Image]],
        return_text: bool = True,
        rec_config: Optional[dict] = None,
        **kwargs,
    ) -> Union[str, List[str], List[Any], List[List[Any]]]:
        """
        Recognize a pure Text Image.
        Args:
            imgs (Union[str, Path, Image.Image], List[str], List[Path], List[Image.Image]): The image or list of images
            return_text (bool): Whether to return only the recognized text; default value is `True`
            rec_config (Optional[dict]): The config for recognition
            kwargs (): Other parameters for `text_ocr.ocr()`

        Returns: Text str or list of text strs when `return_text` is True;
                `List[Any]` or `List[List[Any]]` when `return_text` is False, with the same length as `imgs` and the following keys:

                    * `position`: Position information of the block, `np.ndarray`, with a shape of [4, 2]
                    * `text`: The recognized text
                    * `score`: The confidence score [0, 1]; the higher, the more confident

        """
        is_single_image = False
        if isinstance(imgs, (str, Path, Image.Image)):
            imgs = [imgs]
            is_single_image = True

        input_imgs = prepare_imgs(imgs)

        outs = []
        for image in input_imgs:
            result = self.text_ocr.ocr(np.array(image), rec_config=rec_config, **kwargs)
            if return_text:
                texts = [_one['text'] for _one in result]
                result = '\n'.join(texts)
            outs.append(result)

        if kwargs.get('save_analysis_res'):
            save_layout_img(
                input_imgs[0], ['text'], outs[0], kwargs.get('save_analysis_res'),
            )

        if is_single_image:
            return outs[0]
        return outs

    def recognize_formula(
        self,
        imgs: Union[str, Path, Image.Image, List[str], List[Path], List[Image.Image]],
        batch_size: int = 1,
        return_text: bool = True,
        rec_config: Optional[dict] = None,
        **kwargs,
    ) -> Union[str, List[str], Dict[str, Any], List[Dict[str, Any]]]:
        """
        Recognize pure Math Formula images to LaTeX Expressions
        Args:
            imgs (Union[str, Path, Image.Image, List[str], List[Path], List[Image.Image]): The image or list of images
            batch_size (int): The batch size
            return_text (bool): Whether to return only the recognized text; default value is `True`
            rec_config (Optional[dict]): The config for recognition
            **kwargs (): Special model parameters. Not used for now

        Returns: The LaTeX Expression or list of LaTeX Expressions;
                str or List[str] when `return_text` is True;
                Dict[str, Any] or List[Dict[str, Any]] when `return_text` is False, with the following keys:

                    * `text`: The recognized LaTeX text
                    * `score`: The confidence score [0, 1]; the higher, the more confident

        """
        if self.latex_ocr is None:
            raise RuntimeError('`latex_ocr` model MUST NOT be None')
        outs = self.latex_ocr.recognize(
            imgs, batch_size=batch_size, rec_config=rec_config, **kwargs
        )
        if return_text:
            if isinstance(outs, dict):
                outs = outs['text']
            elif isinstance(outs, list):
                outs = [one['text'] for one in outs]

        return outs


if __name__ == '__main__':
    from .utils import set_logger

    logger = set_logger(log_level='DEBUG')

    p2t = TextFormulaOCR()
    img = 'docs/examples/english.jpg'
    img = read_img(img, return_type='Image')
    out = p2t.recognize(img)
    logger.info(out)
